##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of relstorage.adapters.mysql"""
from __future__ import absolute_import

import logging
import time
import unittest

from ZODB.utils import u64 as bytes8_to_int64
from ZODB.tests import StorageTestBase

from relstorage.adapters.mysql import MySQLAdapter
from relstorage.options import Options
from relstorage._util import timestamp_at_unixtime

from . import StorageCreatingMixin
from . import TestCase

from .util import skipOnCI
from .util import AbstractTestSuiteBuilder
from .util import DEFAULT_DATABASE_SERVER_HOST


class MySQLAdapterMixin(object):

    # The MySQL schema adapter uses DROP TABLE
    # and then CREATE TABLE to zap when ``zap_all(slow=True)``.
    # This is *much* faster than ``DELETE FROM`` on large
    # databases (since we can't use truncate.). But for small databases,
    # it adds lots of extra overhead to re-create those tables all the
    # time, and ``DELETE FROM`` is the way to go.
    zap_slow = True

    def __get_db_name(self):
        if self.keep_history:
            db = self.base_dbname
        else:
            db = self.base_dbname + '_hf'
        return db

    def __get_adapter_options(self, dbname=None):
        dbname = dbname or self.__get_db_name()
        assert isinstance(dbname, str), (dbname, type(dbname))
        return {
            'db': dbname,
            'user': 'relstoragetest',
            'passwd': 'relstoragetest',
            'host': DEFAULT_DATABASE_SERVER_HOST,
        }

    def make_adapter(self, options, db=None):
        return MySQLAdapter(
            options=options,
            **self.__get_adapter_options(db)
        )

    def get_adapter_class(self):
        return MySQLAdapter

    def get_adapter_zconfig(self):
        options = self.__get_adapter_options()
        options['driver'] = self.driver_name
        formatted_options = '\n'.join(
            '     %s %s' % (k, v)
            for k, v in options.items()
        )

        return u"""
        <mysql>
            %s
        </mysql>
        """ % (formatted_options)

    def verify_adapter_from_zconfig(self, adapter):
        self.assertEqual(adapter._params, self.__get_adapter_options())


class TestGenerateTID(MySQLAdapterMixin,
                      StorageCreatingMixin,
                      TestCase,
                      StorageTestBase.StorageTestBase):
    # pylint:disable=too-many-ancestors

    def setUp(self):
        super(TestGenerateTID, self).setUp()
        self._storage = self._closing(self.make_storage())

    def test_extract_parts(self):
        unix_time = 1564063129.1277142

        query = """
        SELECT EXTRACT(year FROM ts) as YEAR,
               EXTRACT(month FROM ts) AS month,
               EXTRACT(day FROM ts) AS day,
               EXTRACT(hour FROM ts) AS hour,
               EXTRACT(minute FROM ts) AS minute,
               %s MOD 60 AS seconds
        FROM (
            SELECT FROM_UNIXTIME(%s) + 0.0 AS ts
        ) t
        """
        cursor = self._storage._load_connection.cursor
        cursor.execute(query, (unix_time, unix_time))
        year, month, day, hour, minute, seconds = cursor.fetchone()
        self.assertEqual(year, 2019)
        self.assertEqual(month, 7)
        self.assertEqual(day, 25)
        self.assertEqual(hour, 13) # If this is not 13, the time_zone is incorrect
        self.assertEqual(minute, 58)
        self.assertEqual(
            round(float(seconds), 6),
            49.127714)

    def test_known_time(self):
        now = 1564054182.277615
        gmtime = (2019, 7, 25, 11, 29, 42, 3, 206, 0)


        self.assertEqual(
            time.gmtime(now),
            gmtime
        )

        ts_now = timestamp_at_unixtime(now)
        self.assertEqual(
            ts_now.raw(),
            b'\x03\xd1Oq\xb4bn\x00'
        )

        self.test_current_time(now)

        # Problematic values due to rounding
        # of minutes due to seconds
        for now, gmtime in (
                (1565774811.9655108,
                 (2019, 8, 14, 9, 26, 51, 2, 226, 0)),
                (1565767799.607957,
                 (2019, 8, 14, 7, 29, 59, 2, 226, 0)),
                (1565775177.915336,
                 (2019, 8, 14, 9, 32, 57, 2, 226, 0)),
                (1565775299.106127,
                 (2019, 8, 14, 9, 34, 59, 2, 226, 0)),
                (1565775479.180209,
                 (2019, 8, 14, 9, 37, 59, 2, 226, 0)),
        ):
            self.assertEqual(time.gmtime(now), gmtime)
            self.test_current_time(now)

    def test_current_time(self, now=None):
        from persistent.timestamp import TimeStamp
        from relstorage._util import int64_to_8bytes
        if now is None:
            now = time.time()
        storage = self._storage
        ts_now = timestamp_at_unixtime(now)

        expected_tid_int = bytes8_to_int64(ts_now.raw())
        __traceback_info__ = now, now % 60.0, time.gmtime(now), ts_now, expected_tid_int

        cursor = storage._load_connection.cursor

        cursor.execute('CALL make_tid_for_epoch(%s, @tid)', (now,))
        cursor.execute('SELECT @tid')
        tid, = cursor.fetchall()[0]

        tid_as_timetime = TimeStamp(int64_to_8bytes(tid)).timeTime()
        __traceback_info__ += (tid_as_timetime - ts_now.timeTime(),)

        self.assertEqual(
            tid,
            expected_tid_int
        )



class MySQLTestSuiteBuilder(AbstractTestSuiteBuilder):

    __name__ = 'MySQL'

    def __init__(self):
        from relstorage.adapters.mysql import drivers
        super(MySQLTestSuiteBuilder, self).__init__(
            drivers,
            MySQLAdapterMixin,
            extra_test_classes=(TestGenerateTID,)
        )

    def _compute_large_blob_size(self, use_small_blobs):
        # MySQL is limited to the blob_chunk_size as there is no
        # native blob streaming support. (Note: this depends on the
        # max_allowed_packet size on the server as well as the driver;
        # both values default to 1MB. So keep it small.)
        return Options().blob_chunk_size

    def _make_check_class_HistoryFreeRelStorageTests(self, bases, name, klass_dict=None):
        bases = (GenericMySQLTestsMixin, ) + bases

        klass_dict = {}

        return self._default_make_check_class(bases, name, klass_dict=klass_dict)

    # pylint:disable=line-too-long
    def _make_check_class_HistoryPreservingRelStorageTests(self, bases, name, klass_dict=None):
        return self._make_check_class_HistoryFreeRelStorageTests(bases, name, klass_dict)


class GenericMySQLTestsMixin(object):

    @skipOnCI("Travis MySQL goes away error 2006")
    def check16MObject(self):
        # NOTE: If your mySQL goes away, check the server's value for
        # `max_allowed_packet`, you probably need to increase it.
        # JAM uses 64M.
        # http://dev.mysql.com/doc/refman/5.7/en/packet-too-large.html
        super(GenericMySQLTestsMixin, self).check16MObject()

    def checkMyISAMTablesProduceErrorWhenNoCreate(self):
        from ZODB.POSException import StorageError
        def cb(_conn, cursor):
            cursor.execute('ALTER TABLE new_oid ENGINE=MyISAM;')
        self._storage._adapter.connmanager.open_and_call(cb)
        # Now open a new storage that's not allowed to create
        with self.assertRaisesRegex(
                StorageError,
                'MyISAM is no longer supported.*new_oid'
        ):
            self.open(create_schema=False)

    def checkMyISAMTablesAutoMigrate(self):
        # Verify we have a broken state.
        self.checkMyISAMTablesProduceErrorWhenNoCreate()
        # Now a storage that can alter a table will do so.
        storage = self.open()
        storage.close()
        storage = self.open(create_schema=False)
        storage.close()

    def checkIsolationLevels(self):

        def assert_storage(storage):
            load_cur = storage._load_connection.cursor

            version_detector = storage._adapter.version_detector
            if not version_detector.supports_transaction_isolation(load_cur):
                raise unittest.SkipTest("Needs MySQL better than %s" % (
                    version_detector.get_version(load_cur)
                ))

            with storage._store_connection_pool.borrowing() as store_connection:
                store_cur = store_connection.cursor
                for cur, ex_iso, ex_ro, ex_timeout in (
                        # Timeout for load is mysql default.
                        [load_cur, 'REPEATABLE-READ', True, 50],
                        [store_cur, 'READ-COMMITTED', False, self.DEFAULT_COMMIT_LOCK_TIMEOUT],
                ):
                    cur.execute("""
                    SELECT @@transaction_isolation,
                           @@transaction_read_only,
                           @@innodb_lock_wait_timeout
                    """)
                    row, = cur.fetchall()
                    iso, ro, timeout = row
                    __traceback_info__ = row
                    iso = iso.decode('ascii') if not isinstance(iso, str) else iso
                    self.assertEqual(iso, ex_iso)
                    self.assertEqual(ro, ex_ro)
                    self.assertEqual(timeout, ex_timeout)

        # By default
        assert_storage(self._storage)

        # In a new instance, and after we do a transaction with it.
        from ZODB.DB import DB
        import transaction

        db = self._closing(DB(self._storage))
        conn = self._closing(db.open())
        assert_storage(conn._storage)

        conn.root()['obj'] = 1
        transaction.commit()

        assert_storage(conn._storage)


def test_suite():
    return MySQLTestSuiteBuilder().test_suite()


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger("zc.lockfile").setLevel(logging.CRITICAL)
    unittest.main(defaultTest="test_suite")



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase
from relstorage.tests import MockOptions

from ..mover import Sqlite3ObjectMover
from ..drivers import Sqlite3Driver
from ...drivers import DriverNotAvailableError

class TestMoverQueries(TestCase):
    maxDiff = None

    def _makeOne(self, keep_history=True):
        try:
            driver = Sqlite3Driver()
        except DriverNotAvailableError as e:
            self.skipTest(e)
        options = MockOptions()
        options.keep_history = keep_history
        return Sqlite3ObjectMover(driver, options)

    def test_detect_conflict_query_hp(self):
        self.assertEqual(
            "SELECT zoid, tid, prev_tid, current_object_state.state "
            "FROM temp_store CROSS JOIN current_object_state USING (zoid) "
            "WHERE (tid <> prev_tid)",
            str(self._makeOne()._detect_conflict_query))

    def test_detect_conflict_query_hf(self):
        self.assertEqual(
            "SELECT zoid, tid, temp_store.prev_tid, object_state.state "
            "FROM temp_store CROSS JOIN object_state USING (zoid) "
            "WHERE (tid <> prev_tid)",
            str(self._makeOne(False)._detect_conflict_query))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for RelStorage

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import shutil
import tempfile

from hamcrest import assert_that
from hamcrest import is_not as does_not
from nti.testing.matchers import provides
from nti.testing.matchers import validly_provides

from ZODB.interfaces import IExternalGC
from ZODB.interfaces import IBlobStorage
from ZODB.interfaces import IBlobStorageRestoreable
from ZODB.interfaces import IStorageUndoable

from relstorage.interfaces import IRelStorage

from relstorage.tests import TestCase
from relstorage.tests import MockAdapter

class TestRelStorage(TestCase):

    def makeOne(self, adapter=None, **kw):
        from relstorage.storage import RelStorage
        # Constructed so as to avoid the need to use a database connection.
        return RelStorage(adapter or MockAdapter(),
                          create=False,
                          cache_prefix='Mock',
                          **kw)

    def test_provides(self):
        storage = self.makeOne()
        assert_that(storage, validly_provides(IRelStorage))
        assert_that(storage, validly_provides(IStorageUndoable))
        assert_that(storage, does_not(provides(IExternalGC)))
        assert_that(storage, does_not(provides(IBlobStorage)))
        assert_that(storage, does_not(provides(IBlobStorageRestoreable)))
        assert_that(storage, does_not(validly_provides(IExternalGC)))

    def test_provides_history_free(self):
        storage = self.makeOne(keep_history=False)
        assert_that(storage, validly_provides(IRelStorage))
        assert_that(storage, does_not(provides(IStorageUndoable)))
        assert_that(storage, does_not(provides(IExternalGC)))
        assert_that(storage, does_not(provides(IBlobStorage)))
        assert_that(storage, does_not(provides(IBlobStorageRestoreable)))
        assert_that(storage, does_not(validly_provides(IExternalGC)))

    def test_provides_external_gc(self):
        adapter = MockAdapter()
        adapter.packundo.deleteObject = True
        storage = self.makeOne(adapter)

        assert_that(storage, validly_provides(IRelStorage))
        assert_that(storage, validly_provides(IExternalGC))

    def test_provides_blob_dir(self):
        tempd = tempfile.mkdtemp(".rstest_storage")
        self.addCleanup(shutil.rmtree, tempd, True)
        storage = self.makeOne(blob_dir=tempd)
        assert_that(storage, validly_provides(IRelStorage))
        assert_that(storage, validly_provides(IBlobStorage))
        assert_that(storage, validly_provides(IBlobStorageRestoreable))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..scriptrunner import ScriptRunner

class Sqlite3ScriptRunner(ScriptRunner):
    script_vars = dict(ScriptRunner.script_vars)
    script_vars.update({
        'TRUNCATE': 'DELETE FROM',
        'FALSE': '0',
        'TRUE': '1'
    })

    def run_script_stmt(self, cursor, generic_stmt, generic_params=()):
        if generic_params and isinstance(generic_params, dict):
            for key in generic_params:
                generic_stmt = generic_stmt.replace('%%(%s)s' % (key,), ':%s' % (key,))

        return ScriptRunner.run_script_stmt(self, cursor, generic_stmt, generic_params)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading


from ZODB.utils import p64

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.storage.tpc.temporary_storage import TemporaryStorage

from ..interfaces import IStorageCache
from . import MockOptionsWithFakeMemcache as MockOptionsWithFakeCache
from . import MockAdapter

class StorageCacheTests(TestCase):
    # pylint:disable=too-many-public-methods

    def setUp(self):
        from relstorage.tests.fakecache import data
        data.clear()
        self._instances = []

    def tearDown(self):
        from relstorage.tests.fakecache import data
        data.clear()
        for inst in self._instances:
            inst.options.cache_local_dir = None
            inst.close()

    def getClass(self):
        from relstorage.cache.storage_cache import StorageCache
        return StorageCache

    def _makeOne(self, current_oids=None, **kwargs):
        options = MockOptionsWithFakeCache.from_args(**kwargs)
        adapter = MockAdapter()
        if current_oids:
            adapter.mover.data.update({oid: (b'', tid) for oid, tid in current_oids.items()})
        inst = self.getClass()(adapter, options,
                               'myprefix')
        self._instances.append(inst)
        inst = inst.new_instance() # coverage and sharing testing

        return inst

    def test_ctor(self):
        from relstorage.tests.fakecache import Client
        from relstorage.cache.memcache_client import MemcacheStateCache
        c = self._makeOne()
        assert_that(c, validly_provides(IStorageCache))

        cache = c.cache
        self.assertIsInstance(cache.g, MemcacheStateCache)
        self.assertIsInstance(cache.g.client, Client)
        self.assertEqual(cache.g.client.servers, ['host:9999'])
        self.assertEqual(c.prefix, 'myprefix')
        self.assertEqual(len(c), 0) # size may be greater than 0
        self.assertEqual(c.limit, MockOptionsWithFakeCache.cache_local_mb * 1000000)

        # can be closed multiple times
        c.close()
        c.close()
        self.test_closed_state(c)

    def test_closed_state(self, c=None):
        if c is None:
            c = self._makeOne()
        c.close()

        self.assertEqual(len(c), 0)
        self.assertTrue(c)
        self.assertEqual(c.size, 0)
        self.assertEqual(c.limit, 0)
        # At no point did we spawn threads.
        self.assertEqual(threading.active_count(), 1)

    def test_stats(self):
        inst = self._makeOne()
        self.assertIsInstance(inst.stats(), dict)
        inst.close()
        self.assertIsInstance(inst.stats(), dict)

    def _setup_for_save(self):
        import tempfile
        import shutil

        c = self._makeOne()
        temp_storage = TemporaryStorage()
        # tid is 2016-09-29 11:35:58,120
        # (That used to matter when we stored that information as a
        # filesystem modification time.)
        tid = 268595726030645777
        oid = 2
        temp_storage.store_temp(oid, b'abc')
        # Flush to the cache.
        c.after_tpc_finish(p64(tid), temp_storage)

        key = list(iter(c.local_client))[0]
        self.assertEqual((2, tid), key)

        c.options.cache_local_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, c.options.cache_local_dir, True)

        # Make like we polled.
        from relstorage.cache import mvcc
        ix = mvcc._ObjectIndex(tid - 1)
        ix = ix.with_polled_changes(tid, tid - 1, [(oid, tid)])
        c.polling_state.object_index = ix
        c.object_index = ix

        return c, oid, tid

    def assertNoPersistentCache(self, cache):
        import os
        from relstorage.cache.persistence import sqlite_files
        fname, _ = sqlite_files(cache.options, cache.prefix)
        if fname:
            self.assertFalse(os.path.exists(fname), fname)

    def assertPersistentCache(self, cache):
        import os
        from relstorage.cache.persistence import sqlite_files
        fname, _ = sqlite_files(cache.options, cache.prefix)
        if fname:
            self.assertTrue(os.path.exists(fname), fname)

    def test_save_and_clear(self):
        c, oid, tid = self._setup_for_save()
        self.assertNoPersistentCache(c)
        c.save(overwrite=True)
        self.assertPersistentCache(c)

        # Creating one in the same place automatically loads it.
        c2 = self._makeOne(current_oids={oid: tid},
                           cache_local_dir=c.options.cache_local_dir)
        self.assertEqual(1, len(c2))

        c.save()

        # Creating a new one loads the stored data.
        c2 = self._makeOne(current_oids={oid: tid},
                           cache_local_dir=c.options.cache_local_dir)
        self.assertEqual(1, len(c2))
        # Invalidating one oid invalidates the polling_state too
        c2.remove_cached_data(oid, tid)
        self.assertEqual(0, len(c2))
        # We can do it again.
        c2.remove_cached_data(oid, tid)
        self.assertEqual(0, len(c2))

        # Resetting also loads the stored data by default.
        c2.clear()
        self.assertEqual(1, len(c2))
        # Invalidating a group of oids invalidates the polling_state too
        c2.remove_all_cached_data_for_oids((oid,))
        self.assertEqual(0, len(c2))
        # We can do it again.
        c2.remove_all_cached_data_for_oids((oid,))
        self.assertEqual(0, len(c2))

        # But can be told to ignore it
        c2.clear(load_persistent=False)
        self.assertEqual(0, len(c2))

        c.options.cache_local_dir = None
        c2.options.cache_local_dir = None

        self.test_closed_state(c2)
        self.test_closed_state(c)

    def test_save_no_hits_no_sets(self):
        c, _, _ = self._setup_for_save()
        c.local_client.reset_stats()
        c.save()
        self.assertNoPersistentCache(c)

    def test_zap_all(self):
        c, _, _ = self._setup_for_save()
        self.assertNoPersistentCache(c)

        c.save(overwrite=True)
        self.assertPersistentCache(c)

        c.zap_all()
        self.assertEmpty(c)
        self.assertNoPersistentCache(c)

        # We can do it again and again
        c.zap_all()
        self.assertEmpty(c)
        self.assertNoPersistentCache(c)

    def test_zap_all_no_local_dir(self):
        c, _, _ = self._setup_for_save()
        self.assertNoPersistentCache(c)
        c.options.cache_local_dir = None

        c.save(overwrite=True)
        self.assertNoPersistentCache(c)

        c.zap_all()
        self.assertEmpty(c)
        self.assertNoPersistentCache(c)

    def test_clear_no_persistent_data(self):
        from relstorage.tests.fakecache import data
        data.clear()
        c = self._makeOne()
        data['x'] = '1'
        c.clear()
        self.assertFalse(data)
        # self.assertEqual(c.checkpoints, None)
        # self.assertEqual(dict(c.delta_after0), {})
        # self.assertEqual(dict(c.delta_after1), {})

    def test_load_without_checkpoints(self):
        c = self._makeOne()
        res = c.load(None, 2)
        self.assertEqual(res, (None, None))

    def test_store_temp(self):
        c = self._makeOne()
        temp_storage = TemporaryStorage()
        temp_storage.store_temp(2, b'abc')
        temp_storage.store_temp(1, b'def')
        temp_storage.store_temp(2, b'ghi')
        self.assertEqual(b'ghi', temp_storage.read_temp(2))
        self.assertEqual(dict(temp_storage.stored_oids),
                         {1: (3, 6, 0), 2: (6, 9, 0)})
        self.assertEqual(temp_storage.max_stored_oid, 2)
        f = temp_storage._queue
        f.seek(0)
        self.assertEqual(f.read(), b'abcdefghi')
        c.after_tpc_finish(p64(3), temp_storage)

        # self.assertEqual(dict(c.delta_after0), {2: 3, 1: 3})

    def test_send_queue_small(self):
        from relstorage.tests.fakecache import data
        c = self._makeOne()
        temp_storage = TemporaryStorage()
        temp_storage.store_temp(2, b'abc')
        temp_storage.store_temp(3, b'def')
        tid = p64(55)
        c.after_tpc_finish(tid, temp_storage)
        self.assertEqual(data, {
            'myprefix:state:55:2': tid + b'abc',
            'myprefix:state:55:3': tid + b'def',
            })
        self.assertEqual(len(c), 2)

    def test_send_queue_large(self):
        from relstorage.tests.fakecache import data
        c = self._makeOne()
        self.assertEqual(c.cache.g.send_limit, 1024 * 1024)
        c.cache.g.send_limit = 100
        temp_storage = TemporaryStorage()
        temp_storage.store_temp(2, b'abc')
        temp_storage.store_temp(3, b'def' * 100)
        tid = p64(55)
        c.after_tpc_finish(tid, temp_storage)
        self.assertEqual(data, {
            'myprefix:state:55:2': tid + b'abc',
            'myprefix:state:55:3': tid + (b'def' * 100),
            })

    def test_send_queue_none(self):
        from relstorage.tests.fakecache import data
        c = self._makeOne()
        temp_storage = TemporaryStorage()
        tid = p64(55)
        c.after_tpc_finish(tid, temp_storage)
        self.assertEqual(data, {})

    def __not_called(self):
        self.fail("Should not be called")

    def test_instances_share_polling_state(self):
        child = self._makeOne()
        self.assertEqual(1, len(self._instances))
        master = self._instances[0]
        self.assertIs(master.polling_state, child.polling_state)

        # This shouldn't actually happen...
        grandchild = child.new_instance()
        self.assertIs(master.polling_state, grandchild.polling_state)
        self.assertTrue(grandchild.polling_state)

        # releasing drops the master
        grandchild.release()
        self.assertFalse(grandchild.polling_state)
        # Doesn't affect anything else.
        self.assertIs(master.polling_state, child.polling_state)



# -*- coding: utf-8 -*-
"""
Tests for the tpc module.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from ZODB.POSException import ReadOnlyError
from ZODB.POSException import StorageTransactionError

from hamcrest import assert_that

from nti.testing.matchers import validly_provides

from ...interfaces import ITPCStateNotInTransaction


class TestNotInTransactionCommon(unittest.TestCase):
    def _makeOne(self, read_only=False, begin_factory=None):
        from .. import NotInTransaction
        return NotInTransaction(begin_factory, read_only)

    def test_verify_interface(self):
        assert_that(self._makeOne(), validly_provides(ITPCStateNotInTransaction))

    def test_is_false(self):
        self.assertFalse(self._makeOne())

    no_transaction_methods = (
        'store', 'restore', 'checkCurrentSerialInTransaction',
        'undo', 'deleteObject', 'restoreBlob',
        'tpc_finish', 'tpc_vote'
    )

    def test_no_transaction(self):
        meths = self.no_transaction_methods
        for meth in meths:
            with self.assertRaises(StorageTransactionError):
                getattr(self._makeOne(), meth)()

class TestNotInTransaction(TestNotInTransactionCommon):

    def test_abort_returns_self(self):
        inst = self._makeOne()
        self.assertIs(inst, inst.tpc_abort())

    def test_initial_state_is_self(self):
        inst = self._makeOne()
        self.assertIs(inst.initial_state, inst)

    def test_read_only_or_no_transaction(self):
        meths = (
            'store',
            'restore',
            'deleteObject',
            'undo',
            'restoreBlob',
        )

        read_only_inst = self._makeOne(read_only=True)
        for meth in meths:
            with self.assertRaises(ReadOnlyError):
                getattr(read_only_inst, meth)()

        writable_inst = self._makeOne(read_only=False)
        for meth in meths:
            with self.assertRaises(StorageTransactionError):
                getattr(writable_inst, meth)()

    def test_with_committed_tid_int(self):
        begin_factory = object()
        read_only = object()
        inst = self._makeOne(read_only, begin_factory)
        self.assertIs(inst.read_only, read_only)
        self.assertIs(inst.begin_factory, begin_factory)
        self.assertEqual(inst.last_committed_tid_int, 0)

        new_inst = inst.with_committed_tid_int(42)
        self.assertIs(new_inst.read_only, read_only)
        self.assertIs(new_inst.begin_factory, begin_factory)
        self.assertEqual(new_inst.last_committed_tid_int, 42)


class TestStale(TestNotInTransactionCommon):

    def _makeOne(self, read_only=False, begin_factory=None):
        from .. import Stale
        from .. import NotInTransaction
        return Stale(NotInTransaction(begin_factory, read_only), StorageTransactionError)

    def test_no_longer_stale(self):
        from .. import NotInTransaction
        inst = self._makeOne()
        self.assertIs(inst.stale(None), inst)
        self.assertIsInstance(inst.no_longer_stale(), NotInTransaction)

    # We piggyback test_no_transaction to mean the stale error
    no_transaction_methods = TestNotInTransactionCommon.no_transaction_methods + (
        'tpc_begin',
    )

class TestSharedTPCState(unittest.TestCase):

    def _makeOne(self, initial_state=None, storage=None, transaction=None):
        from .. import SharedTPCState
        return SharedTPCState(initial_state, storage, transaction)

    class Pool(object):
        replaced = False

        def __init__(self, conn):
            self.conn = conn

        def borrow(self):
            return self.conn

        def replace(self, obj):
            assert obj is self.conn
            self.replaced = True

    class Storage(object):
        def __init__(self, store_pool=None, load_conn=None, blobhelper=None):
            self._store_connection_pool = store_pool
            self._load_connection = load_conn
            self.blobhelper = blobhelper

    def test_store_connection_borrow_release(self):
        from .. import _CLOSED_CONNECTION
        conn = object()

        pool = self.Pool(conn)

        inst = self._makeOne(storage=self.Storage(pool))

        self.assertIs(inst.store_connection, conn)
        inst.release()
        self.assertTrue(pool.replaced)
        self.assertIs(inst.store_connection, _CLOSED_CONNECTION)

    def test_abort_load_connection_store_connection_load_fails(self):
        # Even if we access the load_connection first, the store_connection
        # is aborted first; aborting the store connection doesn't stop the load
        # connection from being aborted.
        from .. import _CLOSED_CONNECTION

        class LoadConn(object):
            aborted = False
            def drop(self):
                self.aborted = True

            exited = False
            def exit_critical_phase(self):
                self.exited = True

        store_conn = object()
        pool = self.Pool(store_conn)
        load_conn = LoadConn()

        storage = self.Storage(pool, load_conn)
        inst = self._makeOne(storage=storage)

        # Order matters
        self.assertIs(inst.load_connection, load_conn)
        self.assertIs(inst.store_connection, store_conn)

        with self.assertRaises(Exception):
            inst.abort(force=True)

        self.assertTrue(load_conn.aborted)
        self.assertTrue(load_conn.exited)
        self.assertTrue(pool.replaced)
        # This got None indicating an exception
        self.assertIsNone(inst.store_connection)
        self.assertIs(inst.load_connection, _CLOSED_CONNECTION)

    def test_blobhelper(self):
        class BlobHelper(object):
            began = False
            aborted = False
            cleared = False

            def begin(self):
                self.began = True
            def abort(self):
                self.aborted = True
            def clear_temp(self):
                self.cleared = True

            txn_has_blobs = True

        blobhelper = BlobHelper()
        storage = self.Storage(blobhelper=blobhelper)
        inst = self._makeOne(storage=storage)
        self.assertFalse(inst.has_blobs())
        self.assertNotIn('blobhelper', vars(inst))

        self.assertIs(inst.blobhelper, blobhelper)
        self.assertTrue(inst.has_blobs())
        self.assertTrue(blobhelper.began)
        inst.abort()
        self.assertIsNone(inst.blobhelper)
        self.assertFalse(inst.has_blobs())
        self.assertTrue(blobhelper.aborted)

        blobhelper = BlobHelper()
        storage = self.Storage(blobhelper=blobhelper)
        inst = self._makeOne(storage=storage)

        self.assertIs(inst.blobhelper, blobhelper)
        inst.release()
        self.assertTrue(blobhelper.cleared)
        self.assertIsNone(inst.blobhelper)

    def test_temp_storage(self):
        # pylint:disable=no-member
        inst = self._makeOne()
        self.assertFalse(inst.has_temp_data())
        self.assertNotIn('temp_storage', vars(inst))

        self.assertIs(inst.temp_storage, inst.temp_storage)
        self.assertIn('temp_storage', vars(inst))
        self.assertFalse(inst.has_temp_data())
        ts = inst.temp_storage
        inst.release()
        self.assertIsNone(ts._queue)
        self.assertIsNone(inst.temp_storage)
        self.assertFalse(inst.has_temp_data())

        inst = self._makeOne()
        ts = inst.temp_storage
        inst.abort()
        self.assertIsNone(ts._queue)
        self.assertIsNone(inst.temp_storage)
        self.assertFalse(inst.has_temp_data())



"""
Packing support for blob data
=============================
"""
from __future__ import absolute_import
from __future__ import print_function

import os

from ZODB.serialize import referencesf
from ZODB.blob import Blob
import transaction

from . import TestBlobMixin

class TestBlobPackHistoryPreservingMixin(TestBlobMixin):
    """
    We need a database with an undoing blob supporting storage.
    """

    def setUp(self):
        super(TestBlobPackHistoryPreservingMixin, self).setUp()
        # The 8 bytes for each TID up to BLOB_REVISION_COUNT
        self.tids = []
        # The timestamps taken just *before* the corresponding
        # TID. These are used as times to pack to.
        self.times = []
        self.oid = None
        self.fns = []
        self._populate()

    #: How many transactions we make with the blob
    BLOB_REVISION_COUNT = 5

    def _populate(self):
        """
        Put some revisions of a blob object in our database and on the
        filesystem.
        """
        from ZODB.utils import u64 as bytes8_to_int64

        connection1 = self.database.open()
        root = connection1.root()

        tids = self.tids = []
        times = self.times = []
        blob = Blob()

        for i in range(self.BLOB_REVISION_COUNT):
            transaction.begin()
            with blob.open('w') as f:
                f.write(b'this is blob data ' + str(i).encode())
            if 'blob' not in root:
                root['blob'] = blob
            transaction.commit()

            blob._p_activate()
            tid = blob._p_serial
            tids.append(tid)
            tid_int = bytes8_to_int64(tid)

            times.append(tid_int - 1)

        blob._p_activate()

        self.oid = oid = root['blob']._p_oid
        fshelper = self.blob_storage.blobhelper.fshelper
        self.fns = [fshelper.getBlobFilename(oid, x) for x in tids]
        connection1.close()

    def _checkFirstCountNotExist(self, count, fns):
        for _ in range(count):
            fn = fns.pop(0)
            self.assertFalse(os.path.exists(fn), fn)
        return fns

    def _checkAllExist(self, fns):
        for f in fns:
            self.assertTrue(os.path.exists(f), f)

    def _pack_at_time_index(self, time_index=0, count_not_exist=0):
        if time_index is None:
            # Use now
            packtime = -1
        else:
            packtime = self.times[time_index]
        self.blob_storage.pack(packtime, referencesf)
        fns = list(self.fns)
        if count_not_exist:
            fns = self._checkFirstCountNotExist(count_not_exist, fns)
        self._checkAllExist(fns)

    def testPack(self):
        # Initially all 5 revisions are present
        self._checkAllExist(self.fns)

        # Packing before the first time changes nothing
        self._pack_at_time_index(0, count_not_exist=0)

        # Ditto for the second
        self._pack_at_time_index(1, count_not_exist=0)

        # Now we start to lose the first one
        self._pack_at_time_index(2, count_not_exist=1)

        # And the second
        self._pack_at_time_index(3, count_not_exist=2)

        # And the third
        self._pack_at_time_index(4, count_not_exist=3)

        # Packing to now leaves only one revision.
        self._pack_at_time_index(None, count_not_exist=4)

        # If we delete the object and do a pack, it should get rid of the most current
        # revision as well as the entire directory:
        dir_name = os.path.split(self.fns[0])[0]
        conn = self.database.open()
        transaction.begin()
        root = conn.root()
        del root['blob']
        transaction.commit()
        conn.close()
        self._pack_at_time_index(None, count_not_exist=5)
        self.assertFalse(os.path.exists(dir_name))

class TestBlobPackHistoryFreeMixin(TestBlobPackHistoryPreservingMixin):
    def _checkFirstCountNotExist(self, count, fns):
        # We don't actually remove anything, until we remove
        # everything; there's really only one file name.
        remaining_fns = fns
        if count == self.BLOB_REVISION_COUNT:
            s = super(TestBlobPackHistoryFreeMixin, self)
            remaining_fns = s._checkFirstCountNotExist(count, fns)
        return remaining_fns



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.tests import MockCursor
from relstorage.tests import MockConnection
from relstorage.tests import MockOptions
from relstorage.tests import MockDriver
from relstorage.tests import DisconnectedException
from relstorage.tests import CloseException

from relstorage.adapters.connmanager import AbstractConnectionManager
from relstorage.adapters import interfaces

class TestAbstractConnectionManager(TestCase):

    def _makeOne(self, **options):
        return AbstractConnectionManager(
            MockOptions.from_args(**options),
            MockDriver()
        )

    def test_provides(self):
        cm = self._makeOne()
        assert_that(cm, validly_provides(interfaces.IConnectionManager))

    def test_close_reports_exception(self):
        cm = self._makeOne()

        self.assertTrue(cm.close(MockConnection()))

        class O(object):
            ex = CloseException
            def close(self):
                raise self.ex

        o = O()
        # As a connection
        self.assertFalse(cm.close(o))
        # As a cursor
        self.assertFalse(cm.close(MockConnection(), o))

        # Raising DisconnectedException instead of CloseException
        o.ex = DisconnectedException
        self.assertFalse(cm.close(o))
        self.assertFalse(cm.close(MockConnection(), o))

    def test_rollback_raises_exception(self):
        cm = self._makeOne()

        # With no exceptions, we should get a success report.
        self.assertTrue(cm.rollback(MockConnection(), MockCursor()))

        class O(object):
            closed = False
            fetched = False
            ex = CloseException
            def rollback(self):
                raise self.ex

            def close(self):
                self.closed = True
                raise CloseException

            def fetchall(self):
                self.fetched = True

        o = O()
        # As a cursor, there's only fetchall(), so nothing is reported.
        conn = MockConnection()
        self.assertTrue(cm.rollback(conn, o))
        self.assertTrue(o.fetched)
        self.assertFalse(o.closed)
        self.assertFalse(conn.closed)

        # As a connection the error is reraised and everything is closed.
        o = O()
        cur = MockCursor()
        cur.fetchall = lambda: setattr(cur, 'fetched', True)
        with self.assertRaises(o.ex):
            cm.rollback(o, cur)

        self.assertTrue(o.closed)
        self.assertTrue(cur.fetched) # pylint:disable=no-member
        self.assertTrue(cur.closed)

        # Raising DisconnectedException instead of CloseException
        o = O()
        o.ex = DisconnectedException
        conn = MockConnection()
        self.assertTrue(cm.rollback(conn, o))
        self.assertTrue(o.fetched)
        self.assertFalse(o.closed)
        self.assertFalse(conn.closed)

        o = O()
        o.ex = DisconnectedException
        cur = MockCursor()
        with self.assertRaises(o.ex):
            cm.rollback(o, cur)

        self.assertTrue(o.closed)
        self.assertTrue(cur.closed)

    def test_rollback_and_close_reports_exception(self):
        cm = self._makeOne()

        # With no exceptions, we should get a success report.
        self.assertTrue(cm.rollback_and_close(MockConnection(), MockCursor()))

        class O(object):
            closed = False
            fetched = False
            ex = CloseException
            def rollback(self):
                raise self.ex

            def close(self):
                self.closed = True
                raise CloseException

            def fetchall(self):
                self.fetched = True

        o = O()
        # As a cursor, we fetchall, then we close, so we get a report.
        self.assertFalse(cm.rollback_and_close(MockConnection(), o))
        self.assertTrue(o.fetched)
        self.assertTrue(o.closed)

        o = O()
        # As a connection the error is reported and everything is closed.
        self.assertFalse(cm.rollback_and_close(o, MockCursor()))
        self.assertTrue(o.closed)

        # Raising DisconnectedException instead of CloseException
        o = O()
        o.ex = DisconnectedException
        self.assertFalse(cm.rollback_and_close(MockConnection(), o))
        self.assertTrue(o.fetched)
        self.assertTrue(o.closed)

        o = O()
        o.ex = DisconnectedException
        self.assertFalse(cm.rollback_and_close(o, MockCursor()))
        self.assertTrue(o.closed)

    def test_rollback_quietly_reports_exception(self):
        cm = self._makeOne()
        meth = cm.rollback_quietly
        # With no exceptions, we should get a success report.
        self.assertTrue(meth(MockConnection(), MockCursor()))

        class O(object):
            closed = False
            fetched = False
            ex = CloseException
            def rollback(self):
                raise self.ex

            def close(self):
                self.closed = True
                raise CloseException

            def fetchall(self):
                self.fetched = True

        o = O()
        # As a cursor, we fetchall, but don't close, so we get no report.
        self.assertTrue(meth(MockConnection(), o))
        self.assertTrue(o.fetched)
        self.assertFalse(o.closed)

        o = O()
        # As a connection the error is reported and everything is closed.
        self.assertFalse(meth(o, MockCursor()))
        self.assertTrue(o.closed)

        # Raising DisconnectedException instead of CloseException
        o = O()
        o.ex = DisconnectedException
        self.assertTrue(meth(MockConnection(), o))
        self.assertTrue(o.fetched)
        self.assertFalse(o.closed)

        o = O()
        o.ex = DisconnectedException
        self.assertFalse(meth(o, MockCursor()))
        self.assertTrue(o.closed)

    def test_without_replica_conf(self):
        cm = self._makeOne()

        conn = MockConnection()
        cm.restart_load(conn, MockCursor())
        self.assertTrue(conn.rolled_back)

        conn = MockConnection()
        cm.restart_store(conn, MockCursor())
        self.assertTrue(conn.rolled_back)

    def test_with_replica_conf(self):
        import os
        import relstorage.tests
        tests_dir = relstorage.tests.__file__
        replica_conf = os.path.join(os.path.dirname(tests_dir), 'replicas.conf')

        cm = self._makeOne(replica_conf=replica_conf)

        conn = MockConnection()
        conn.replica = '127.0.0.1'
        cm.restart_load(conn, MockCursor())
        self.assertTrue(conn.rolled_back)
        conn.replica = 'other'
        self.assertRaises(interfaces.ReplicaClosedException,
                          cm.restart_load, conn, MockCursor())

        conn = MockConnection()
        conn.replica = '127.0.0.1'
        cm.restart_store(conn, MockCursor())
        self.assertTrue(conn.rolled_back)
        conn.replica = 'other'
        self.assertRaises(interfaces.ReplicaClosedException,
                          cm.restart_store, conn, MockCursor())

    def test_with_ro_replica_conf(self):
        import os
        import relstorage.tests
        tests_dir = relstorage.tests.__file__
        replica_conf = os.path.join(os.path.dirname(tests_dir),
                                    'replicas.conf')
        ro_replica_conf = os.path.join(os.path.dirname(tests_dir),
                                       'ro_replicas.conf')

        cm = self._makeOne(replica_conf=replica_conf, ro_replica_conf=ro_replica_conf)

        conn = MockConnection()
        conn.replica = 'readonlyhost'
        cm.restart_load(conn, MockCursor())
        self.assertTrue(conn.rolled_back)
        conn.replica = 'other'
        self.assertRaises(interfaces.ReplicaClosedException,
                          cm.restart_load, conn, MockCursor())



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase

from relstorage.cache import interfaces
from relstorage.cache import _statecache_wrappers as wrappers

class MockCache(object):
    released = False

    def new_instance(self):
        return type(self)()

    def release(self):
        self.released = True

class TestMultiStateCache(TestCase):

    def _makeOne(self, l=None, g=None):
        return wrappers.MultiStateCache(l, g)

    def test_provides(self):
        assert_that(self._makeOne(),
                    validly_provides(interfaces.IStateCache))

    def test_delitem_found(self):
        c = self._makeOne({(1, 1): None}, {(1, 1): None})
        del c[(1, 1)]

    def test_new_instance_and_release(self):
        root = self._makeOne(MockCache(), MockCache())
        child = root.new_instance()
        self.assertIsNot(child.l, root.l)
        self.assertIsNot(child.g, root.g)

        child.release()
        self.assertIsNone(child.l)
        self.assertIsNone(child.g)

class MockTracer(object):
    def __init__(self):
        self.trace_events = []

    def trace(self, code, oid_int=0, tid_int=0, end_tid_int=0, dlen=0):
        self.trace_events.append((code, oid_int, tid_int, end_tid_int, dlen))

    def trace_store_current(self, tid_int, state_oid_iter):
        """Does nothing"""

    def close(self):
        self.trace_events = tuple(self.trace_events)

class TestTracingCacheWrapper(TestCase):

    def _makeOne(self, kind=None):
        kind = kind or wrappers.MultiStateCache

        return wrappers.TracingStateCache(
            kind(MockCache(), MockCache()),
            MockTracer())

    def test_provides(self):
        assert_that(self._makeOne(),
                    validly_provides(interfaces.IStateCache))
        assert_that(self._makeOne().new_instance(),
                    validly_provides(interfaces.IStateCache))

    def test_new_instance_and_release(self):
        root = self._makeOne()
        child = root.new_instance()
        self.assertIs(root.tracer, child.tracer)
        child_cache = child.cache
        child_cache_l = child.cache.l
        child_cache_g = child.cache.g

        child.release()

        self.assertIsNotNone(child.tracer)
        self.assertIsNone(child.cache)
        self.assertIsNone(child_cache.l)
        self.assertIsNone(child_cache.g)
        self.assertTrue(child_cache_l.released)
        self.assertTrue(child_cache_g.released)
        self.assertFalse(root.cache.l.released)
        self.assertFalse(root.cache.g.released)

        child.close()
        self.assertIsNone(child.tracer)

    def test_get_miss(self):
        from . import LocalClient
        c = self._makeOne(lambda *args: LocalClient(100))

        self.assertIsNone(c[(1, 0)])

        self.assertEqual(
            c.tracer.trace_events,
            [(0x20, 1, 0, 0, 0,)]
        )

    def test_get_hit(self):
        from . import LocalClient
        c = self._makeOne(lambda *args: LocalClient(100))
        c.cache[(1, 1)] = (b'abc', 1)
        self.assertEqual(c[(1, 1)], (b'abc', 1))

        self.assertEqual(
            c.tracer.trace_events,
            [(0x22, 1, 1, 0, 3,)]
        )



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Locker implementations.
"""

from __future__ import absolute_import
from __future__ import print_function

import abc
import sys
import six


from relstorage._compat import ABC
from relstorage._compat import metricmethod
from relstorage._util import consume
from relstorage._util import Lazy

from ._util import query_property as _query_property
from ._util import DatabaseHelpersMixin
from .schema import Schema

from .interfaces import UnableToAcquireCommitLockError
from .interfaces import UnableToLockRowsToModifyError
from .interfaces import UnableToLockRowsToReadCurrentError

logger = __import__('logging').getLogger(__name__)

class AbstractLocker(DatabaseHelpersMixin,
                     ABC):

    def __init__(self, options, driver, batcher_factory):
        self.keep_history = options.keep_history
        self.commit_lock_timeout = options.commit_lock_timeout
        self.commit_lock_id = options.commit_lock_id
        self.driver = driver
        self.lock_exceptions = driver.lock_exceptions
        self.illegal_operation_exceptions = driver.illegal_operation_exceptions
        self.make_batcher = batcher_factory

    def on_store_opened(self, cursor, restart=False):
        """
        A callback that must be called when a store connection is
        opened or restarted.

        This implementation calls :meth:`_on_store_opened_set_row_lock_timeout`
        when the store connection is initially opened.
        """
        self._on_store_opened_set_row_lock_timeout(cursor, restart)

    def _on_store_opened_set_row_lock_timeout(self, cursor, restart=False):
        """
        Install a per-session row lock timeout.

        This should be set to the :attr:`commit_lock_timeout`.

        This applies to the locks taken by :meth:`lock_current_objects` during the
        ``tpc_vote`` phase of a transaction.
        """

    #: Set this to a false value if you don't support ``NOWAIT`` and we'll
    #: instead set the lock timeout to a low value.
    supports_row_lock_nowait = True

    def _set_row_lock_timeout(self, cursor, timeout):
        "Implement this to change the row lock timeout."

    def _set_row_lock_nowait(self, cursor): # pragma: no cover
        """
        If :attr:`supports_row_lock_timeout` is not true, then this
        class will call this method when :meth:`hold_commit_lock` is
        called with a false value for *nowait*.

        This method in turn is implemented to call
        :meth:`_set_row_lock_timeout` with a *timeout* argument of 0.

        This method is deprecated, as the *nowait* parameter is deprecated.

        ..note::

            This class does not attempt to revert this change, leaving
            the connection in a 0 timeout.

            Previously, we only set *nowait* to false during packing, and only
            for that one connection. So it didn't matter that it wasn't rolled back.

            Now, we don't use *nowait* at all.

            We could automatically make this change local to the transaction
            on PostgreSQL with ``SET LOCAL``, and on MySQL with the
            optimizer hint ``SET_VAR()`` in the lock statement --- but only in 8+.
        """
        self._set_row_lock_timeout(cursor, 0)

    _lock_current_clause = 'FOR UPDATE'
    _lock_share_clause_nowait = 'FOR SHARE NOWAIT'
    _lock_share_clause = 'FOR SHARE'

    #: These double as the query to get OIDs we'd like to lock, but
    #: do not actually lock them.
    _get_current_objects_queries = (
        # If we also include the objects being added,
        # mysql takes out gap locks, and we can deadlock?
        # TODO: Confirm.
        ("""
        SELECT zoid
        FROM current_object
        WHERE zoid IN (
            SELECT zoid
            FROM temp_store
        )
        """, 'current_object'),
        ("""
        SELECT zoid
        FROM object_state
        WHERE zoid IN (
            SELECT zoid
            FROM temp_store
        )
        """, 'object_state'),
    )

    _get_current_objects_query = _query_property('_get_current_objects')

    @Lazy
    def _lock_current_objects_query(self):
        return '{select} ORDER BY ZOID {lock}'.format(
            select=self._get_current_objects_query[0],
            lock=self._lock_current_clause
        )

    @metricmethod
    def lock_current_objects(self, cursor, read_current_oid_ints, shared_locks_block,
                             after_lock_share=lambda: None):
        # We need to be sure to take the locks in a deterministic
        # order; the easiest way to do that is to order them by OID.
        # But we have two separate sets of OIDs we need to lock: the
        # ones we're finding the current data for, and the ones that
        # we're going to check for conflicts. The ones we're checking
        # for conflicts are already in the database in ``temp_store``
        # (and partly in the storage cache's temporary storage and/or
        # the row batcher); the current oids are only in memory.
        #
        # So we have a few choices: either put the current oids into
        # a database table and do a UNION query with temp_store,
        # or pull the temp_store data into memory, union it with
        # current_oids and issue a single big query.
        #
        # Our strategy could even vary depending on the size of
        # *current_oids*; in the usual case, it will be small or
        # empty, and an in-database big UNION query is probably
        # workable (in the empty case, we can and do elide this part
        # altogether)


        # In history free mode, *table* will be `object_state`, which
        # has ZOID as its primary key. In history preserving mode,
        # *table* will be `current_object`, where ZOID is also the primary
        # key (and `object_state` is immutable; that's why we don't need
        # to take any locks there --- conflict resolution will always be able
        # to find the desired state without fear of it changing).

        # MySQL 8 allows ``NOWAIT`` and ``SKIP LOCKED`` on ``FOR
        # UPDATE`` or ``FOR SHARE`` clauses; earlier versions do not
        # have that. PostgreSQL allows both.

        # In all databases, the locks we get depend on the indexing.
        # We must be searching on the primary key to get the smallest,
        # most specific row locks. In some databases, rows are only
        # locked when they are returned by the cursor, so we must
        # consume all the rows.

        # Lock order is explained in IRelStorageAdapter.lock_objects_and_detect_conflicts.

        # First lock rows we need shared access to, typically without blocking.
        if read_current_oid_ints:
            self._lock_readCurrent_oids_for_share(cursor, read_current_oid_ints,
                                                  shared_locks_block)
            after_lock_share()


        # Then lock the rows we need exclusive access to.
        # This will block for up to ``commit_lock_timeout``.
        self._lock_rows_being_modified(cursor)

    def _lock_readCurrent_oids_for_share(self, cursor, current_oids, shared_locks_block):
        _, table = self._get_current_objects_query
        oids_to_lock = sorted(set(current_oids))
        batcher = self.make_batcher(cursor)

        locking_suffix = ' %s ' % (
            self._lock_share_clause
            if shared_locks_block
            else
            self._lock_share_clause_nowait
        )
        try:
            rows = batcher.select_from(
                ('zoid',), table,
                suffix=locking_suffix,
                **{'zoid': oids_to_lock}
            )
            consume(rows)
        except self.illegal_operation_exceptions: # pragma: no cover
            # Bug in our code
            raise
        except self.lock_exceptions:
            self.reraise_commit_lock_error(
                cursor,
                'SELECT zoid FROM {table} WHERE zoid IN () {lock}'.format(
                    table=table,
                    lock=locking_suffix
                ),
                UnableToLockRowsToReadCurrentError
            )

    def _lock_rows_being_modified(self, cursor):
        stmt = self._lock_current_objects_query
        try:
            cursor.execute(stmt)
            rows = cursor
            consume(rows)
        except self.illegal_operation_exceptions: # pragma: no cover
            # Bug in our code
            raise
        except self.lock_exceptions:
            self.reraise_commit_lock_error(
                cursor,
                stmt,
                UnableToLockRowsToModifyError
            )

    def reraise_commit_lock_error(self, cursor, lock_stmt, kind):
        try:
            debug_info = self._get_commit_lock_debug_info(cursor, True)
        except Exception as nested: # pylint:disable=broad-except
            logger.exception("Failed to get lock debug info")
            debug_info = "%r(%r)" % (type(nested), nested)
        __traceback_info__ = lock_stmt, debug_info
        if debug_info:
            logger.debug("Failed to acquire commit lock:\n%s", debug_info)
        message = "Acquiring a lock during commit failed: %s%s" % (
            sys.exc_info()[1],
            '\n' + debug_info if debug_info else '(No debug info.)'
        )
        six.reraise(
            kind,
            kind(message),
            sys.exc_info()[2])

    # MySQL allows aggregates in the top level to use FOR UPDATE,
    # but PostgreSQL does not, so we have to use the second form.
    #
    # 'SELECT MAX(tid) FROM transaction FOR UPDATE',
    # 'SELECT tid FROM transaction WHERE tid = (SELECT MAX(tid) FROM transaction)  FOR UPDATE',

    # Note that using transaction in history-preserving databases
    # can still lead to deadlock in older versions of MySQL (test
    # checkPackWhileWriting), and the above lock statement can
    # lead to duplicate transaction ids being inserted on older
    # versions (5.7.12, PyMySQL:
    # https://ci.appveyor.com/project/jamadden/relstorage/builds/25748619/job/cyio3w54uqi026lr#L923).
    # So both HF and HP use an artificial lock row.
    #
    # TODO: Figure out exactly the best way to lock just the rows
    # in the transaction table we care about that works
    # everywhere, or a better way to choose the next TID.
    # gap/intention locks might be a clue.
    _commit_lock_query = Schema.commit_row_lock.select(
        Schema.commit_row_lock.c.tid
    ).for_update(
    ).prepared(
    )

    _commit_lock_nowait_query = _commit_lock_query.nowait()

    @metricmethod
    def hold_commit_lock(self, cursor, ensure_current=False, nowait=False):
        # pylint:disable=unused-argument
        #
        # This method is not used for PostgreSQL or MySQL.
        lock_stmt = self._commit_lock_query
        if nowait: # pragma: no cover
            if self.supports_row_lock_nowait:
                lock_stmt = self._commit_lock_nowait_query
            else:
                self._set_row_lock_nowait(cursor)

        try:
            lock_stmt.execute(cursor)
            rows = cursor.fetchall()
            if not rows or not rows[0]:
                raise UnableToAcquireCommitLockError("No row returned from commit_row_lock")
        except self.illegal_operation_exceptions: # pragma: no cover
            # Bug in our code.
            raise
        except self.lock_exceptions:
            if nowait:
                return False
            self.reraise_commit_lock_error(
                cursor,
                lock_stmt,
                UnableToAcquireCommitLockError
            )

        return True

    def _get_commit_lock_debug_info(self, cursor, was_failure=False):
        # pylint:disable=unused-argument
        """
        Subclasses can implement this to return a string
        that will be added to the exception message when a commit lock cannot
        be acquired. For example, it might list other connections that
        have conflicting locks.
        """
        return ''

    @abc.abstractmethod
    def release_commit_lock(self, cursor):
        raise NotImplementedError()

    @abc.abstractmethod
    def hold_pack_lock(self, cursor):
        raise NotImplementedError()

    @abc.abstractmethod
    def release_pack_lock(self, cursor):
        raise NotImplementedError()



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Oracle adapter for RelStorage."""
from __future__ import absolute_import

import logging

from zope.interface import implementer

from ..adapter import AbstractAdapter
from ..dbiter import HistoryFreeDatabaseIterator
from ..dbiter import HistoryPreservingDatabaseIterator
from ..interfaces import IRelStorageAdapter
from ..poller import Poller

from . import drivers
from .batch import OracleRowBatcher
from .connmanager import CXOracleConnectionManager
from .locker import OracleLocker
from .mover import OracleObjectMover
from .oidallocator import OracleOIDAllocator
from .packundo import OracleHistoryFreePackUndo
from .packundo import OracleHistoryPreservingPackUndo
from .schema import OracleSchemaInstaller
from .scriptrunner import CXOracleScriptRunner
from .stats import OracleStats
from .txncontrol import OracleTransactionControl

log = logging.getLogger(__name__)


@implementer(IRelStorageAdapter)
class OracleAdapter(AbstractAdapter):
    """Oracle adapter for RelStorage."""
    # pylint:disable=too-many-instance-attributes

    driver_options = drivers

    def __init__(self,
                 user='relstoragetest',
                 password='relstoragetest',
                 dsn='192.168.1.131/orcl',
                 commit_lock_id=0,
                 twophase=False, options=None,
                 locker=None,
                 mover=None,
                 connmanager=None,
                 ):
        """Create an Oracle adapter.

        The user, password, and dsn parameters are provided to
        cx_Oracle at connection time.

        If twophase is true, all commits go through an Oracle-level
        two-phase commit process. This is disabled by default; it causes
        deadlock detection to become slow and based on a timeout. Even
        when this option is disabled, the ZODB two-phase commit is
        still in effect.
        """
        # pylint:disable=unused-argument
        self._user = user
        self._password = password
        self._dsn = dsn
        self._twophase = twophase
        self.locker = locker
        self.mover = mover
        self.connmanager = connmanager

        super(OracleAdapter, self).__init__(options)

    def _create(self):
        driver = self.driver
        user = self._user
        password = self._password
        twophase = self._twophase
        options = self.options
        dsn = self._dsn

        batcher_factory = lambda cursor, row_limit=None: OracleRowBatcher(
            cursor, inputsizes, row_limit
        )
        if self.connmanager is None:
            self.connmanager = CXOracleConnectionManager(
                driver,
                user=user,
                password=password,
                dsn=dsn,
                twophase=twophase,
                options=options,
            )
        self.runner = CXOracleScriptRunner(driver)
        if self.locker is None:
            self.locker = OracleLocker(
                options=self.options,
                driver=driver,
                inputsize_NUMBER=driver.NUMBER,
                batcher_factory=batcher_factory,
            )
        self.schema = OracleSchemaInstaller(
            connmanager=self.connmanager,
            runner=self.runner,
            keep_history=self.keep_history,
        )
        if self.mover is None:
            inputsizes = {
                'blobdata': driver.BLOB,
                'rawdata': driver.BINARY,
                'oid': driver.NUMBER,
                'tid': driver.NUMBER,
                'prev_tid': driver.NUMBER,
                'chunk_num': driver.NUMBER,
                'md5sum': driver.STRING,
            }
            self.mover = OracleObjectMover(
                driver,
                options=options,
                runner=self.runner,
                batcher_factory=batcher_factory,
            )
            self.mover.inputsizes = inputsizes

        self.oidallocator = OracleOIDAllocator(
            connmanager=self.connmanager,
        )

        self.poller = Poller(
            self.driver,
            keep_history=self.keep_history,
            runner=self.runner,
            revert_when_stale=options.revert_when_stale,
            transactions_may_go_backwards=(
                self.connmanager.replica_selector is not None
                or self.connmanager.ro_replica_selector is not None
            )
        )

        self.txncontrol = OracleTransactionControl(
            connmanager=self.connmanager,
            poller=self.poller,
            keep_history=self.keep_history,
            Binary=driver.Binary,
            twophase=twophase,
        )



        if self.keep_history:
            self.packundo = OracleHistoryPreservingPackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryPreservingDatabaseIterator(
                driver,
            )
        else:
            self.packundo = OracleHistoryFreePackUndo(
                driver,
                connmanager=self.connmanager,
                runner=self.runner,
                locker=self.locker,
                options=options,
            )
            self.dbiter = HistoryFreeDatabaseIterator(
                driver,
            )

        self.stats = OracleStats(
            connmanager=self.connmanager,
            keep_history=self.keep_history
        )

    def new_instance(self):
        # This adapter and its components are stateless, so it's
        # safe to share it between threads.
        return OracleAdapter(
            user=self._user,
            password=self._password,
            dsn=self._dsn,
            twophase=self._twophase,
            options=self.options,
            locker=self.locker,
            mover=self.mover,
            connmanager=self.connmanager,
        )

    def __str__(self):
        parts = [self.__class__.__name__]
        if self.keep_history:
            parts.append('history preserving')
        else:
            parts.append('history free')
        parts.append('user=%r' % self._user)
        parts.append('dsn=%r' % self._dsn)
        parts.append('twophase=%r' % self._twophase)
        return ", ".join(parts)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint:disable=unused-argument
import os
import os.path
import random
import time
from collections import namedtuple

from pyperf import perf_counter

from relstorage.options import Options
from relstorage._util import get_memory_usage
from relstorage._util import byte_display

logger = __import__('logging').getLogger(__name__)

try:
    import sys
    import cProfile
    import pstats
    if '--profile' not in sys.argv:
        raise ImportError
except ImportError:
    class cProfile(object):
        class Profile(object):
            def enable(self):
                pass
            def disable(self):
                pass
    class pstats(object):
        class Stats(object):
            def __init__(self, *args):
                pass
            def sort_stats(self, *args):
                return self
            def print_stats(self, *args):
                pass


class MockOptions(Options):
    cache_module_name = ''
    cache_servers = ''
    cache_local_object_max = 16384
    cache_local_compression = 'none'
    cache_local_dir_count = 1

def profiled(func, name, runner):
    import vmprof # pylint:disable=import-error
    import tempfile


    def f(*args):
        prefix = name.replace(' ', '_') + '-'
        suffix = '.vmprof' + "%d%d" % (sys.version_info[:2])
        handle, _ = tempfile.mkstemp(suffix, prefix=prefix, dir=runner.args.temp)
        vmprof.enable(handle, lines=True)
        result = func(*args)
        vmprof.disable()
        os.close(handle)
        return result

    return f

def _combine_benchmark_results(options, group_name, benchmark_group, cache_options):
    # Do this in the master only, after running all the benchmarks
    # for a database.
    if options.worker:
        return

    if options.output:
        from pyperf import Benchmark
        from pyperf import BenchmarkSuite

        # Create a file for the entire suite, using names that can
        # be compared across different database configurations.
        dir_name = os.path.splitext(options.output)[0] + '.d'
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        # We're going to update the metadata, so we need to make
        # a copy.
        # Use the short name so that even runs across different object
        # counts are comparable.
        for name, benchmark in list(benchmark_group.items()):
            benchmark = Benchmark(benchmark.get_runs())
            benchmark.update_metadata({'name': name})
            benchmark_group[name] = benchmark
        suite = BenchmarkSuite(benchmark_group.values())

        fname = os.path.join(dir_name,
                             group_name + '_' + str(cache_options.cache_local_mb) + '.json')
        suite.dump(fname, replace=True)


def run_and_report_funcs(runner, named_funcs):
    profile = runner.args.profile
    benchmarks = {}
    for description in named_funcs:
        func = description[1]
        name = description[0]
        args = description[2:]
        if profile:
            func = profiled(func, name, runner)
        benchmark = runner.bench_time_func(name, func, *args)
        benchmarks[name] = benchmark
    return benchmarks

def _read_random(data_size):
    with open('/dev/urandom', 'rb') as f:
        return f.read(data_size)


def _make_key_groups(key_group_size):

    key_groups = []
    key_groups.append(list(range(key_group_size)))
    for i in range(key_group_size):
        keys = list(range(key_group_size * i, key_group_size * (i + 1)))
        assert len(set(keys)) == len(keys)
        key_groups.append(keys)

    return key_groups

def _make_data(data_or_size, key_group_size):
    random_data = None
    if isinstance(data_or_size, bytes):
        random_data = data_or_size
    else:
        random_data = _read_random(data_or_size)

    key_groups = _make_key_groups(key_group_size)
    # Recent PyPy and Python 3.6 preserves iteration order of a dict
    # to match insertion order. If we use a dict for ALL_DATA, this
    # gives slightly different results due to the key lengths being
    # different and so things being ejected at slightly different
    # times (on PyPy, 8 key groups have *no* matches in read() using a dict,
    # while that doesn't occur in cPython 2.7/3.4). To
    # make this all line up the same, we preserve order everywhere by using
    # a list of tuples (sure enough, that change makes 8 groups go missing)
    # Alternately, if we sort by the hash of the key, we get the iteration order that
    # CPython used for a dict, making all groups of keys be found in read(). This
    # keeps the benchmark consistent

    ALL_DATA = {}
    for group in key_groups:
        for key in group:
            tid = key
            ALL_DATA[key] = (random_data, tid)
    ALL_DATA = list(ALL_DATA.items())
    ALL_DATA.sort(key=lambda x: hash(x[0]))
    return ALL_DATA


def local_benchmark(runner):
    # pylint:disable=too-many-statements,too-many-locals
    import gc
    from relstorage.cache.local_client import LocalClient

    options = MockOptions()
    options.cache_local_mb = 100
    options.cache_local_compression = 'none'

    KEY_GROUP_SIZE = 400
    DATA_SIZE = 1024

    # With 1000 in a key group, and 1024 bytes of data, we produce
    # 909100 keys, and 930918400 = 887MB of data, which will overflow
    # a cache of 500 MB.

    # A group size of 100 produces 9100 keys with 9318400 = 8.8MB of data.
    # Likewise, group of 200 produces 36380 keys with 35.5MB of data.

    # Group size of 400 produces 145480 keys with 142MB of data.

    # Most of our time is spent in compression, it seems.
    # In the 8.8mb case, populating all the data with default compression
    # takes about 2.5-2.8s. Using no compression, it takes 0.38 to 0.42s.
    # Reading is the same at about 0.2s.
    random_data = _read_random(DATA_SIZE)

    # To best capture the actual memory usage, only create the data
    # as needed, and destroy it before taking a memory reading.

    def makeOne(populate=True, data=None):
        mem_before = get_memory_usage()
        gc.collect()
        gc.collect()
        objects_before = len(gc.get_objects())

        client = LocalClient(options, 'pfx')
        client.b_mem_before = mem_before
        client.b_objects_before = objects_before
        if populate:
            client._bulk_update([
                (t[0], (t[1][0], t[1][1], False, 1))
                for t
                in data or _make_data(random_data, KEY_GROUP_SIZE)
            ])
        return client

    def report(client, duration, extra=''):
        gc.collect()
        gc.collect()
        mem_before = client.b_mem_before
        objects_before = client.b_objects_before
        mem_after = get_memory_usage()
        mem_used = mem_after - mem_before
        objects_after = len(gc.get_objects())
        logger.info(
            "%s:%sExecuted in %s; Mem delta: %s;  Object Delta: %s; "
            "Num keys: %d; Weight: %s; "
            "Mem before: %s; Mem after: %s",
            os.getpid(), extra,
            duration, byte_display(mem_used), objects_after - objects_before,
            len(client), byte_display(client.size),
            byte_display(mem_before), byte_display(mem_after),
        )

    def populate_equal(loops):
        all_data = _make_data(random_data, KEY_GROUP_SIZE)
        client = makeOne(populate=False)

        duration = 0

        for _ in range(loops):
            begin = perf_counter()
            for oid, (state, tid) in all_data:
                # install the same version again.
                # this should mean no extra copies,
                # net even temporary
                client[(oid, tid)] = (state, tid)
            duration += perf_counter() - begin

            report(client, duration, extra=" (Loop " + str(_) + ") ")

        report(client, duration, extra=" (Final ) ")
        return duration

    def populate_not_equal(loops):
        # Because we will populate when we make,
        # capture memory now to be able to include that.
        client = makeOne()
        duration = 0

        for loop in range(loops):
            all_data = _make_data(random_data, KEY_GROUP_SIZE)

            begin = perf_counter()
            for oid, (state, tid) in all_data:
                # install a copy that's not quite equal.
                # This should require saving it.
                # Note that we must use a different TID, or we
                # get CacheConsistencyError.
                state = state + str(loop).encode('ascii')
                tid = tid + loop + 1
                key = (oid, tid)
                new_v = (state, tid)
                client[key] = new_v
            duration += perf_counter() - begin
            all_data = None
            report(client, duration, extra=" (Loop " + str(loop) + ") ")

        report(client, duration, extra=" (Final ) ")
        return duration

    def populate_empty(loops):
        # Capture memory usage
        init_client = makeOne(populate=False)
        # Here, we keep all clients alive so that we can
        # measure the total memory usage.
        all_clients = []
        duration = 0
        all_data = _make_data(random_data, KEY_GROUP_SIZE)
        for loop in range(loops):
            client = makeOne(populate=False)
            all_clients.append(client)
            begin = perf_counter()
            for oid, (state, tid) in all_data:
                client[(oid, tid)] = (state, tid)
            duration += perf_counter() - begin
            all_data = None
            report(client, duration, extra=" (Loop " + str(loop) + ") ")

        report(init_client, duration, extra=" (Final ) ")
        return duration

    def populate_bulk(loops):
        # Capture memory usage
        all_data = _make_data(random_data, KEY_GROUP_SIZE)
        kvs = [
            (t[0], (t[1][0], t[1][1], False, 1))
            for t
            in all_data
        ]

        init_client = makeOne(populate=False)
        duration = 0
        for loop in range(loops):
            client = makeOne(populate=False)
            begin = perf_counter()
            client._bulk_update(kvs)
            duration += perf_counter() - begin
            report(client, duration, extra=" (Loop " + str(loop) + ") ")
            client = None

        report(init_client, duration, extra=" (Final ) ")
        return duration

    def read(loops):
        # This is basically the worst-case scenario for a basic
        # segmented LRU: A repeating sequential scan, where no new
        # keys are added and all existing keys fit in the two parts of the
        # cache. Thus, entries just keep bouncing back and forth between
        # probation and protected. It so happens that this is our slowest
        # case.
        client = makeOne(populate=True)
        key_groups = _make_key_groups(KEY_GROUP_SIZE)
        begin = perf_counter()
        duration = 0

        for _ in range(loops):
            client = makeOne(populate=True)
            begin = perf_counter()
            for keys in key_groups:
                for oid in keys:
                    tid = oid
                    key = (oid, tid)
                    res = client[key]
                    #assert len(res) == len(keys)
                    if not res:
                        continue
                    assert res[0] == random_data
            duration += (perf_counter() - begin)
        print("Hit ratio: ", client.stats()['ratio'])

        duration = perf_counter() - begin
        key_groups = None

        report(client, duration, extra=" (Final ) ")
        return duration
        # import pprint
        # pprint.pprint(client._bucket0.stats())
        # print("Probation promotes", client._bucket0._probation.promote_count)
        # print("Probation demotes", client._bucket0._probation.demote_count)
        # print("Probation removes", client._bucket0._probation.remove_count)

    def mixed(loops):
        key_groups = _make_key_groups(KEY_GROUP_SIZE)
        duration = 0
        hot_keys = key_groups[0]

        client = makeOne(populate=True)
        i = 0
        miss_count = 0

        for _ in range(loops):
            all_data = _make_data(random_data, KEY_GROUP_SIZE)

            begin = perf_counter()
            for oid, (state, tid) in all_data:
                i += 1
                key = (oid, tid)
                client[key] = (state, tid)
                if i == len(hot_keys):
                    for hot_oid in hot_keys:
                        hot_key = (hot_oid, hot_oid)
                        res = client[hot_key]
                        if not res:
                            miss_count += 1
                    i = 0
            duration += perf_counter() - begin
            all_data = None

        report(client, duration, extra="(Final)")
        return duration
    # def mixed_for_stats():
    #     # This is a trivial function that simulates the way
    #     # new keys can come in over time as we reset our checkpoints.
    #     # (Actually, it mostly shows our superiority over the plain LRU;
    #     # that one scored a 0.0 hit ratio, where our segmented LRU scores 1.0)
    #     client.reset_stats()
    #     hot_keys = key_groups[0]
    #     i = 0
    #     for _k, v in ALL_DATA:
    #         i += 1
    #         client._bucket0[str(i)] = v


    #     client.get_multi(hot_keys)

    #     print("Hit ratio", client.stats()['ratio'])

    groups = {}
    for name in ('CFFI',):
        benchmarks = run_and_report_funcs(
            runner,
            (
                (name + ' pop_bulk', populate_bulk),
                (name + ' pop_eq', populate_equal),
                (name + ' pop_ne', populate_not_equal),
                (name + ' epop', populate_empty),
                (name + ' read', read),
                (name + ' mix ', mixed),
            ))
        group = {
            k[len(name) + 1:]: v for k, v in benchmarks.items()
        }
        groups[name] = group

    if not runner.args.worker:
        for name, group in groups.items():
            _combine_benchmark_results(runner.args, name, group, options)


StorageRecord = namedtuple('Record', ['asu', 'lba', 'size', 'opcode', 'ts'])

class StorageTraceSimulator(object):
    # Text trace files can be obtained from http://traces.cs.umass.edu/index.php/Storage/Storage
    # Binary trace files can be obtained from https://github.com/cache2k/cache2k-benchmark

    def _open_file(self, filename, mode='r'):
        if filename.endswith('.bz2'):
            import bz2
            f = bz2.BZ2File(filename, mode)
        else:
            f = open(filename, mode)
        return f

    def _read_binary_records(self, filename, num_clients=8, write_pct=.30,
                             mean_size=10000, stddev_size=512):
        # pylint:disable=too-many-locals
        import struct
        keys = []
        i = 0
        with self._open_file(filename, 'rb') as f:
            while True:
                key = f.read(4)
                if not key:
                    break
                key = struct.unpack(">I", key)[0]
                key = str(key)
                keys.append((i, key))
                i += 1

        random.seed("read_binary_records")
        write_keys = set(random.sample(keys, int(len(keys) * write_pct)))

        records = []
        for key in keys:
            size = int(random.normalvariate(mean_size, stddev_size))
            opcode = 'r'
            if key in write_keys:
                opcode = 'w'
            asu = 1 if num_clients == 1 else random.randrange(num_clients)

            records.append(StorageRecord(asu, key[1], size, opcode, 0.0))
        return records

    def _read_text_records(self, filename):
        from relstorage._compat import intern as _intern

        records = []
        with self._open_file(filename) as f:
            for line in f:
                if isinstance(line, bytes) and str is not bytes:
                    line = line.decode('ascii')
                fields = [x.strip() for x in line.split(",")]
                fields[0] = int(fields[0]) # asu
                try:
                    fields[1] = _intern(fields[1]) # lba
                    fields[2] = int(fields[2]) # size
                    fields[3] = _intern(fields[3].lower()) # opcode
                    fields[4] = float(fields[4]) # ts
                except IndexError:
                    print("Invalid line", line)
                    continue

                records.append(StorageRecord(*fields[:5]))

        return records

    def read_records(self, filename):
        if filename.endswith(".trace"):
            return self._read_binary_records(filename)
        return self._read_text_records(filename)

    def _report_one(self, stats, f, cache_local_mb, begin_time, end_time):
        stats['time'] = end_time - begin_time
        print("{:15s} {:>5s} {:>7s} {:>7s} {:>5s}".format(
            "File", "Limit", "Size", "Time", "Hits"))
        print("{:15s} {:5d} {:7.2f} {:7.2f} {:.3f}".format(
            os.path.basename(f), cache_local_mb,
            stats['bytes'] / 1024 / 1024, stats['time'],
            stats['ratio']))

    def _simulate_local(self, records, cache_local_mb, f):
        from relstorage.cache.local_client import LocalClient
        options = MockOptions()
        options.cache_local_mb = cache_local_mb
        options.cache_local_compression = 'none'
        client = LocalClient(options)

        now = time.time()
        for record in records:
            key = record.lba

            if record.opcode == 'r':
                data = client[key]
                if data is None:
                    # Fill it in from the backend
                    client[key] = b'r' * record.size
            else:
                assert record.opcode == 'w'
                client[key] = b'x' * record.size

        done = time.time()
        stats = client.stats()
        self._report_one(stats, f, cache_local_mb, now, done)

        return stats

    def _simulate_storage(self, records, cache_local_mb, f):
        # pylint:disable=too-many-locals,too-many-statements
        from relstorage.cache.storage_cache import StorageCache
        from relstorage.cache.tests import MockAdapter
        from ZODB.utils import p64

        TRANSACTION_SIZE = 10

        options = MockOptions()
        options.cache_local_mb = cache_local_mb
        options.cache_local_compression = 'none'
        #options.cache_delta_size_limit = 30000
        adapter = MockAdapter()

        # Populate the backend with data, all as of tid 1 Use the size
        # for the first time we see the data, just like
        # _simulate_local does. If we choose a small fixed size, we
        # get much better hit rates than _simulate_local If we use the
        # actual size of the first time we see each record, we use an
        # insane amount of memory even interning the strings
        # (WebSearch3 requires 12GB of memory), so we create just the biggest value
        # and then take memoryviews of it to avoid any copies.

        max_size = 0
        first_sizes = {}
        for record in records:
            max_size = max(record.size, max_size)
            if record.lba in first_sizes:
                continue
            first_sizes[record.lba] = record.size

        # Create one very big value, and then use subviews of a memoryview to reference
        # the same memory

        max_size = max(first_sizes.values())
        biggest_value = b'i' * max_size
        biggest_value = memoryview(biggest_value)

        for lba, size in first_sizes.items():
            oid = int(lba)
            adapter.mover.data[oid] = (biggest_value[:size], 1)
            assert len(adapter.mover.data[oid][0]) == size
        assert len(adapter.mover.data) == len(first_sizes)
        root_cache = StorageCache(adapter, options, None)

        if '--store-trace' in sys.argv:
            class RecordingCache(object):

                def __init__(self, cache):
                    self._cache = cache
                    self.operations = []

                def set_multi(self, data):
                    for k, v in data.items():
                        self.operations.append(('w', k, len(v)))
                    return self._cache.set_multi(data)

                def set(self, k, v):
                    self.operations.append(('w', k, len(v)))
                    self._cache[k] = v

                def get(self, k):
                    self.operations.append(('r', k, -1))
                    return self._cache.get(k)

                def get_multi(self, keys):
                    for k in keys:
                        self.operations.append(('r', k, -1))
                    return self._cache.get_multi(keys)

                def stats(self):
                    return self._cache.stats()

            local_client = root_cache.local_client
            local_client = RecordingCache(local_client)
            root_cache.local_client = root_cache.cache = local_client

        # Initialize to the current TID
        current_tid_int = 2
        root_cache.polling_state.object_index.maps[0].highest_visible_tid = current_tid_int

        # Each ASU is a connection, so it has its own storage cache instance.
        asu_caches = {asu: root_cache.new_instance()
                      for asu
                      in set((x.asu for x in records))}

        for cache in asu_caches.values():
            cache.after_poll(None, 0, current_tid_int, [])
            cache.bm_current_tid = current_tid_int
            cache.bm_changes = {}

        now = time.time()

        for record in records:
            oid_int = int(record.lba)
            cache = asu_caches[record.asu]

            # Poll after a certain number of operations, or of we know we would get a
            # conflict.
            if (current_tid_int - cache.bm_current_tid >= TRANSACTION_SIZE
                    or oid_int in cache.bm_changes):
                cache.after_poll(None, cache.bm_current_tid, current_tid_int,
                                 cache.bm_changes.items())
                cache.bm_current_tid = current_tid_int
                cache.bm_changes.clear()

            if record.opcode == 'r':
                cache.load(None, oid_int)
            else:
                assert record.opcode == 'w'
                current_tid_int += 1
                cache.tpc_begin()
                new_state = biggest_value[:record.size]
                cache.store_temp(oid_int, new_state)
                adapter.mover.data[oid_int] = (new_state, current_tid_int)
                cache.after_tpc_finish(p64(current_tid_int))

                for cache in asu_caches.values():
                    cache.bm_changes[oid_int] = current_tid_int

        done = time.time()
        stats = root_cache.stats()
        self._report_one(stats, f, cache_local_mb, now, done)

        if hasattr(root_cache.local_client, 'operations'):
            with open(f + '.' + str(options.cache_local_mb) + '.ctrace', 'w') as fp:
                for o in root_cache.local_client.operations:
                    fp.write("%s,%s,%d\n" % o)
        return stats

    def simulate(self, s_type='local'):
        meth = getattr(self, '_simulate_' + s_type)

        def _print(size, records):
            print("Simulating", len(records),
                  "operations (reads:", (len([x for x in records if x.opcode == 'r'])),
                  "writes:", (len([x for x in records if x.opcode == 'w'])), ")",
                  "to", len(set(x.lba for x in records)), "distinct keys",
                  "from", len(set((x.asu for x in records))), "connections",
                  "with cache limit", size)

        filename = sys.argv[2]
        filename = os.path.abspath(os.path.expanduser(filename))
        if os.path.isdir(filename):
            all_stats = []
            for f in sorted(os.listdir(filename)):
                records = self.read_records(os.path.join(filename, f))
                for size in (100, 512, 1024):
                    _print(size, records)
                    stats = meth(records, size, f)
                    all_stats.append((f, size, stats))

            print("{:15s} {:>5s} {:>7s} {:>7s} {:>5s}".format(
                "File", "Limit", "Size", "Time", "Hits"))
            for f, size, stats in all_stats:
                print("{:15s} {:5d} {:7.2f} {:7.2f} {:.3f}".format(
                    os.path.basename(f), size, stats['bytes'] / 1024 / 1024,
                    stats['time'], stats['ratio']))

        else:
            size = int(sys.argv[3])
            records = self.read_records(filename)
            _print(size, records)
            pr = cProfile.Profile()
            pr.enable()
            meth(records, size, filename)
            pr.disable()
            ps = pstats.Stats(pr).sort_stats('cumulative')
            ps.print_stats(.4)


def save_load_benchmark(runner):
    # pylint:disable=too-many-locals,too-many-statements
    import io
    from relstorage.cache import persistence as _Loader
    from relstorage.cache.local_client import LocalClient

    import itertools

    sys.setrecursionlimit(500000)


    cache_pfx = "pfx"
    cache_options = MockOptions()
    cache_options.cache_local_dir = runner.args.temp #'/tmp'
    cache_options.cache_local_dir_compress = False
    cache_options.cache_local_mb = 525

    # Note use of worker task number: This is fragile and directly relates to
    # the order in which we pass functions to run_and_report_funcs
    def create_and_populate_client():
        client = LocalClient(cache_options, cache_pfx)
        # Monkey with the internals so we don't have to
        # populate multiple times.
        #print("Testing", type(bucket._dict))

        # Only need to populate in the workers.
        size_dists = [100] * 800 + [300] * 500 + [1024] * 300 + [2048] * 200 + [4096] * 150

        with open('/dev/urandom', 'rb') as rnd:
            data = [rnd.read(x) for x in size_dists]
        data_iter = itertools.cycle(data)
        keys_and_values = []
        len_values = 0
        j = 0
        for j, datum in enumerate(data_iter):
            if len(datum) > client.limit or len_values + len(datum) > client.limit:
                break
            len_values += len(datum)
            # To ensure the pickle memo cache doesn't just write out "use object X",
            # but distinct copies of the strings, we need to copy them
            keys_and_values.append(
                ((j, j), (datum[:-1] + b'x', j))
            )

            # # We need to get the item so its frequency goes up enough to be written
            # # (this is while we're doing an aging at write time, which may go away).
            # # Using an assert statement causes us to write nothing if -O is used.
            # if bucket[(j, j)] is datum:
            #     raise AssertionError()
        mem_before = get_memory_usage()
        client._bulk_update(keys_and_values, mem_usage_before=mem_before)
        del keys_and_values
        #print("Len", len(bucket), "size", bucket.size, "checkpoints", client.get_checkpoints())
        return client

    def _open(fd, mode):
        return io.open(fd, mode, buffering=16384)

    def write_client():
        client = create_and_populate_client()
        begin = perf_counter()
        client.save(overwrite=True)
        end = perf_counter()
        return end - begin

    def write_client_dups():
        client = create_and_populate_client()
        begin = perf_counter()
        client.save(overwrite=False, close_async=False)
        end = perf_counter()
        return end - begin

    def read_client():
        begin = perf_counter()
        c2 = LocalClient(cache_options, cache_pfx)
        c2.restore()
        end = perf_counter()
        return end - begin

    benchmarks = run_and_report_funcs(runner, (
        ('write client fresh', write_client),
        ('write client dups', write_client_dups),
        ('read client', read_client),
    ))

    if not runner.args.worker:
        stream = {'write': benchmarks['write stream'],
                  'read': benchmarks['read stream']}
        db = {'write': benchmarks['write client fresh'],
              'read': benchmarks['read client']}
        db_dups = {'write': benchmarks['write client dups'],
                   'read': benchmarks['read client']}
        _combine_benchmark_results(runner.args, 'stream', stream, cache_options)
        _combine_benchmark_results(runner.args, 'db', db, cache_options)
        _combine_benchmark_results(runner.args, 'db_dup', db_dups, cache_options)


def main():
    import pyperf
    import tempfile
    # NOTE: If you're trying to use clang's address sanitizer
    # on these, at least on macOS, the necessary DLYD_ environment
    # variable doesn't get passed to the subprocess; it's not anything that
    # --inherit-environ can fix. You have to hard code it here *and* add it
    # to --inherit-environ.
    temp = None

    def args_hook(cmd, args):
        cmd.extend(('--type', args.type))
        if args.profile:
            cmd.extend(('--profile',))
        if args.do_stream:
            cmd.extend(('--do-stream',))
        if args.log:
            cmd.extend(('--log',))
        cmd.extend(('--temp', temp))
    runner = pyperf.Runner(add_cmdline_args=args_hook)
    runner.argparser.add_argument(
        '--type',
        default='io',
        choices=['local', 'io', 'simlocal', 'simstorage']
    )
    runner.argparser.add_argument(
        '--temp'
    )
    runner.argparser.add_argument(
        '--profile', action='store_true',
    )
    runner.argparser.add_argument(
        '--log', action='store_true'
    )
    runner.argparser.add_argument(
        '--do-stream', action='store_true',
    )
    runner.parse_args()
    kind = runner.args.type
    temp = runner.args.temp
    need_cleanup = False
    if not temp:
        temp = tempfile.mkdtemp('.rsbench')
        need_cleanup = True
    if runner.args.log:
        import logging
        logging.basicConfig(level=logging.DEBUG)

    if kind == 'local':
        local_benchmark(runner)
    elif kind == 'io':
        save_load_benchmark(runner)
    elif kind == 'simlocal':
        StorageTraceSimulator().simulate('local')
    else:
        assert kind == 'simstorage'
        #import logging
        #logging.basicConfig(level=logging.DEBUG)

        StorageTraceSimulator().simulate('storage')

    if not runner.args.worker and need_cleanup:
        # Master process cleanup
        import shutil
        shutil.rmtree(temp)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase
from relstorage.tests import MockCursor

from ..schema import MySQLVersionDetector as MVD

class TestMySQLVersionDetector(TestCase):

    def test_5(self):
        mvd = MVD()
        cursor = MockCursor()
        cursor.results.append((b'5.7.27-log',))
        vi = mvd.get_version_info(cursor)
        self.assertEqual(vi, (5, 7, 27))
        self.assertEqual(mvd.get_major_version(cursor), 5)
        self.assertFalse(mvd.supports_nowait(None))
        self.assertEqual(mvd.get_version(cursor), '5.7.27-log')
        self.assertGreater(mvd.get_version_info(cursor), (5, 7, 26))
        self.assertGreater(mvd.get_version_info(cursor), (5, 7, 20))
        self.assertTrue(mvd.supports_good_stored_procs(None))

        self.assertEqual([('SELECT version()', None)], cursor.executed)

    def test_8(self):
        mvd = MVD()
        cursor = MockCursor()
        cursor.results.append((b'8.0.19-standard',))
        vi = mvd.get_version_info(cursor)
        self.assertEqual(vi, (8, 0, 19))
        self.assertEqual(mvd.get_major_version(cursor), 8)
        self.assertTrue(mvd.supports_nowait(None))
        self.assertTrue(mvd.supports_transaction_isolation(None))
        self.assertEqual(mvd.get_version(cursor), '8.0.19-standard')
        self.assertGreater(mvd.get_version_info(cursor), (5, 7, 20))
        self.assertEqual([('SELECT version()', None)], cursor.executed)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Interfaces for top-level RelStorage components.

These interfaces aren't meant to be considered public, they exist to
serve as documentation and for validation of RelStorage internals.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import Interface
from zope.interface import Attribute

# pylint:disable=inherit-non-class, no-self-argument, no-method-argument
# pylint:disable=too-many-ancestors

class IBlobHelper(Interface):
    """
    Blob support for RelStorage.

    There is one `IBlobHelper` per storage instance. Each
    `IBlobHelper` instance has access to the associated adapter as
    well as shared instances of ``fshelper`` (a
    ``ZODB.blob.FilesystemHelper``) and ``cache_checker`` (a
    BlobCacheChecker).
    """

    NEEDS_DB_LOCK_TO_FINISH = Attribute("Boolean")
    NEEDS_DB_LOCK_TO_VOTE = Attribute("Boolean")

    def new_instance(adapter):
        """
        Create a new instance for use in a new MVCC storage.
        """

    ###
    # Reading
    ###

    def loadBlob(cursor, oid, serial):
        pass

    def openCommittedBlobFile(cursor, oid, serial, blob=None):
        pass

    ###
    # Writing.
    #
    # This is only valid to do after a call to :meth:`begin`.
    ###

    def temporaryDirectory():
        pass

    def storeBlob(cursor, store_func,
                  oid, serial, data, blobfilename, version, txn):
        """Storage API: store a blob object."""


    ###
    # Transactions
    ###

    txn_has_blobs = Attribute("Does the transaction this object is joined to include blobs?")

    def begin():
        """
        Start a new transaction.
        """

    def vote(tid=None):
        """
        Check the transaction can be committed.

        If the *tid* is None, meaning it hasn't been allocated yet,
        then, if this implementation requires a TID in order to vote,
        it may raise an `StorageTransactionError`. If that happens,
        lock the database, allocated a TId, and try again.

        As an implementation note, does nothing *unless* we have a
        shared blob dir.
        """

    def finish(tid):
        """
        Finalize the transaction.

        As an implementation note, does nothing *if* we have a shared
        blob dir.
        """

    def abort():
        """
        Abort the transaction.
        """

    ###
    # Undo
    ###

    def copy_undone(copied, tid):
        """
        After an undo operation, copy the matching blobs forward.

        The copied parameter is a list of ``(integer oid, integer tid)``.

        Does nothing if not a ``shared_blob_dir``.
        """

    def restoreBlob(cursor, oid, serial, blobfilename):
        pass

    ###
    # Misc
    ###
    def after_pack(oid_int, tid_int):
        """
        Called after an object state has been removed by packing.

        Removes the corresponding blob file.
        """

    def close():
        pass


class IAuthoritativeBlobHelper(IBlobHelper):
    """
    A blob helper that has the only copy of the blob data.
    """

class ICachedBlobHelper(IBlobHelper):
    """
    A blob helper that is only a cache; the real data is elsewhere.
    """


class INoBlobHelper(IBlobHelper):
    """
    An object that does nothing with blobs.

    Used to avoid conditional logic in the main code. Methods that
    impact the use of the storage (user tries to store a blob but
    that's not possible, etc) should raise an error. Methods that are
    part of the internal workings of the storage and would have no
    side-effects (because there cannot be blobs) should quietly do nothing.
    """



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Tests for config.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from relstorage import config
from relstorage import options

from . import mock

class MockZConfig(object):

    def __init__(self, section_name):
        self.__section_name = section_name

    def getSectionName(self):
        return self.__section_name

    def getSectionAttributes(self):
        return (k for k in self.__dict__ if not k.startswith('_'))

class MockStorage(MockZConfig):
    adapter = None
    name = 'name'

class MockAdapter(object):

    def __init__(self):
        self.config = MockZConfig('adapter')
        self.config.driver = None
        self.options = None

    def create(self, opts):
        self.options = opts

    def __eq__(self, other):
        return isinstance(other, MockAdapter)

class TestConfig(unittest.TestCase):

    DRIVER_NAME = 'a_driver'

    def _make_expected_options(self, **kwargs):
        o = options.Options(name='name',
                            driver=self.DRIVER_NAME)
        for k, v in kwargs.items():
            setattr(o, k, v)
        return o

    @mock.patch('relstorage.config.RelStorage', autospec=True)
    def test_driver_stays_set(self, mock_storage):
        # There was a bug where opening a configuration object more than
        # once would lose the driver settings.
        root_zconfig = MockStorage('relstorage')
        root_zconfig.adapter = MockAdapter()
        root_zconfig.adapter.config.driver = self.DRIVER_NAME

        factory = config.RelStorageFactory(root_zconfig)

        factory.open()

        mock_storage.assert_called_once_with(
            None,
            name='name',
            options=self._make_expected_options(adapter=MockAdapter()))

        # The driver was copied to the right place to make an Options instance
        self.assertEqual(root_zconfig.driver, self.DRIVER_NAME)
        # And it was left where it was
        self.assertEqual(root_zconfig.adapter.config.driver, self.DRIVER_NAME)

        # It was in the options
        self.assertEqual(root_zconfig.adapter.options.driver, self.DRIVER_NAME)

    @mock.patch('relstorage.adapters.mysql.MySQLAdapter')
    @mock.patch('relstorage.config.RelStorage', autospec=True)
    def test_mysql_doesnt_pass_driver_as_connect_kwarg(self, _, mock_mysql):

        root_zconfig = MockStorage('relstorage')
        adapter_zconfig = MockZConfig('adapter')
        adapter_zconfig.driver = self.DRIVER_NAME
        adapter_zconfig.a_conn_setting = 42 # pylint:disable=attribute-defined-outside-init
        root_zconfig.adapter = config.MySQLAdapterFactory(adapter_zconfig)


        factory = config.RelStorageFactory(root_zconfig)

        factory.open()

        mock_mysql.assert_called_once_with(
            a_conn_setting=42,
            options=self._make_expected_options(adapter=root_zconfig.adapter))


class TestOptions(unittest.TestCase):

    def test_deprecated_warnings(self):
        import warnings
        for name in options.Options._deprecated_options:
            with warnings.catch_warnings(record=True) as caught:
                warnings.simplefilter('default')
                class C(object):
                    pass
                setattr(C, name, 42)
                options.Options.copy_valid_options(C)

            self.assertEqual(1, len(caught))
            self.assertIn(name, str(caught[0].message))



##############################################################################
#
# Copyright (c) 2009,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import os
import time

import ZODB.blob

from ZODB.POSException import StorageTransactionError

from relstorage._compat import iteritems
from relstorage._compat import MAC

from .util import lock_blob

class AbstractBlobHelper(object):
    """
    Stores blobs on the filesystem. This base class
    handles everything that doesn't depend on whether or
    not the disk storage is canonical (`shared_blob_dir`)
    or just a cache.
    """

    # _txn_blobs: {oid_bytes->filename}; contains blob data for the
    # currently uncommitted transaction.
    _txn_blobs = None

    #: A ZODB.blob.FilesystemHelper object. Subclasses must create.
    fshelper = None

    def __init__(self, options, adapter, fshelper):
        self.options = options
        self.adapter = adapter
        self.blob_dir = options.blob_dir
        self.fshelper = fshelper
        self.new_instance_kwargs = {}

    def __repr__(self):
        return "<%s at 0x%x blob_dir=%r txn_blobs=%s>" % (
            type(self).__name__,
            id(self),
            self.blob_dir,
            len(self._txn_blobs) if self._txn_blobs is not None else None
        )

    def new_instance(self, adapter):
        return type(self)(
            self.options,
            adapter,
            self.fshelper,
            **self.new_instance_kwargs
        )

    def clear_temp(self):
        self._txn_blobs = None

    def begin(self):
        if self._txn_blobs is not None:
            raise StorageTransactionError("Already in a transaction.")
        self._txn_blobs = {}

    @property
    def txn_has_blobs(self):
        return bool(self._txn_blobs)

    def close(self):
        self._txn_blobs = None
        self.fshelper = None
        self.options = None
        self.adapter = None
        self.new_instance_kwargs = None

    def _get_lockable_blob_filename(self, oid, serial):
        # Create the directory if needed
        self.fshelper.getPathForOID(oid, True)
        return self.fshelper.getBlobFilename(oid, serial)

    def _lock_blob_for_download(self, oid, serial):
        blob_filename = self._get_lockable_blob_filename(oid, serial)
        return lock_blob(blob_filename)

    if MAC:
        _lock_blob_for_open = _lock_blob_for_download
    else:
        _lock_blob_for_open = lambda self, oid, serial: None

    def loadBlob(self, cursor, oid, serial):
        return self._loadBlobInternal(cursor, oid, serial)

    def _loadBlobInternal(self, cursor, oid, serial, blob_lock=None):
        raise NotImplementedError

    @staticmethod
    def _accessed(filename):
        try:
            os.utime(filename, (time.time(), os.stat(filename).st_mtime))
        except OSError:
            pass # We tried. :)
        return filename

    def _cachedLoadBlobInternal(self, oid, serial):
        # Load a blob.
        # Note that the thread that cleans the blob cache up when it reaches
        # a maximum size could remove the blob file by the time the caller
        # gets the filename, so it could be gone.
        blob_filename = self.fshelper.getBlobFilename(oid, serial)
        if os.path.exists(blob_filename):
            return self._accessed(blob_filename)

    def _openCommittedBlobFileInternal(self, cursor, oid, serial, blob, open_lock):
        blob_filename = self._loadBlobInternal(cursor, oid, serial, open_lock)
        if blob is None:
            result = open(blob_filename, 'rb')
        else:
            result = ZODB.blob.BlobFile(blob_filename, 'r', blob)
        return result

    def openCommittedBlobFile(self, cursor, oid, serial, blob=None):
        # First, try to make sure the file exists on disk.
        #
        # Next, open and return it. This would be expected to either
        # return a file we can read, or raise a FileNotFoundError.
        # Sadly, on some platforms (macOS 10.14.5 with APFS on Python
        # 3), this has a race condition with a concurrent unlink
        # syscall from the cache cleaner, such that the open succeeds,
        # but reading fails to return any data, depending on how the
        # open and unlink syscalls are interleaved. So we must be sure
        # to prevent the two from overlapping; we do that by holding
        # the lock. See https://github.com/zodb/relstorage/issues/219
        #
        # Unfortunately, but not unexpectedly, this about doubles the
        # amount of time it takes to open blobs that are already
        # present. So we jump through some hoops to only do this on
        # platforms that we know need it.
        blob_lock = self._lock_blob_for_open(oid, serial)
        try:
            try:
                return self._openCommittedBlobFileInternal(cursor, oid, serial, blob, blob_lock)
            except IOError:
                # An IOError here should mean that the file couldn't
                # be opened, probably because the cache cleaner came
                # through and deleted it. If we had already opened a
                # lock, then there's nothing we can do (the cache
                # cleaner wouldn't have been able to delete it).
                # However, we do test that we retry in that case.
                if blob_lock is None:
                    blob_lock = self._lock_blob_for_download(oid, serial)
                # If we didn't have the lock, we need to try again with the lock.
                return self._openCommittedBlobFileInternal(cursor, oid, serial, blob, blob_lock)
        finally:
            if blob_lock is not None:
                blob_lock.close()

    def temporaryDirectory(self):
        return self.fshelper.temp_dir

    def _doStoreBlob(self, store_func,
                     oid, serial, data, blobfilename, txn):
        """Storage API: store a blob object."""
        # Grab the file right away. That way, if we don't have enough
        # room for a copy, we'll know now rather than in tpc_finish.
        # Also, this relieves the client of having to manage the file
        # (or the directory contianing it).
        self.fshelper.getPathForOID(oid, create=True)
        fd, temp_path = self.fshelper.blob_mkstemp(oid, serial)
        os.close(fd)

        # It's a bit odd (and impossible on windows) to rename over an
        # existing file. We'll use the temporary file name as a base.
        temp_path += '-'
        ZODB.blob.rename_or_copy_blob(blobfilename, temp_path)
        os.remove(temp_path[:-1])
        self._add_blob_to_transaction(oid, temp_path)

        store_func(oid, serial, data, txn)
        return temp_path

    def _add_blob_to_transaction(self, oid, filename):
        old_filename = self._txn_blobs.get(oid)
        if old_filename is not None and old_filename != filename:
            ZODB.blob.remove_committed(old_filename)
        self._txn_blobs[oid] = filename

    def _move_blobs_into_place(self, tid):
        if not self._txn_blobs:
            return 0
        if not tid:
            raise StorageTransactionError("No TID for blobs")
        assert isinstance(tid, bytes)
        # We now have a transaction ID, so rename all the blobs
        # accordingly. This is very unlikely to fail. If we're
        # not using a shared blob-dir, it doesn't matter much if it fails;
        # source data is safely in the database, we'd just have some extra temporary
        # files. (Though we don't want that exception to populate from tpc_finish.)
        #
        # In fact, ClientStorage does this in tpc_finish for blob cache dirs.
        # It's not been reported as a problem there, so probably it really does
        # rarely fail. Exceptions from tpc_finish are a VERY BAD THING.
        total_size = 0
        for oid, sourcename in self._txn_blobs.items():
            size = os.stat(sourcename).st_size
            total_size += size
            targetname = self.fshelper.getBlobFilename(oid, tid)
            if sourcename != targetname:
                lock = lock_blob(targetname)
                try:
                    ZODB.blob.rename_or_copy_blob(sourcename, targetname)
                finally:
                    lock.close()
                self._txn_blobs[oid] = targetname
        return total_size

    def vote(self, tid=None):
        """
        Does nothing.
        """

    def finish(self, tid): # pylint:disable=unused-argument
        """
        Ends the transaction. Subclasses must call.
        """
        self.clear_temp()

    def _abort_filename(self, filename):
        """
        Needs to do something for shared blob dirs.
        """

    def abort(self):
        try:
            if not self._txn_blobs:
                return

            for _oid, filename in iteritems(self._txn_blobs):
                if os.path.exists(filename):
                    ZODB.blob.remove_committed(filename)
                    self._abort_filename(filename)
        finally:
            self.clear_temp()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import itertools
import sys
import time
import traceback
import os

import logging
from logging import DEBUG
from logging import INFO
from logging import WARN
from logging import ERROR

from persistent.timestamp import TimeStamp

from ZConfig.datatypes import asBoolean
from ZConfig.datatypes import integer
from ZConfig.datatypes import RangeCheckedConversion
from ZConfig.datatypes import stock_datatypes

from ZODB.utils import p64
from ZODB.utils import u64
# Trace is beneath (less important than) DEBUG.
# It is "extremely verbose"
from ZODB.loglevels import TRACE

from relstorage._compat import wraps
from relstorage._compat import update_wrapper
from relstorage._compat import perf_counter

logger = logging.getLogger(__name__)
perf_logger = logger.getChild('timing')

int64_to_8bytes = p64
bytes8_to_int64 = u64

__all__ = [
    'int64_to_8bytes',
    'bytes8_to_int64',
    'timestamp_at_unixtime',
    'timer',
    'log_timed',
    'log_timed_only_self',
    'thread_spawn',
    'spawn',
    'get_memory_usage',
    'byte_display',
    'Lazy',
    'CachedIn',
    'to_utf8',
    'consume',
    'TRACE',
    'parse_boolean',
    'parse_byte_size',
    'positive_integer',
    'get_time_from_environ',
]

positive_integer = RangeCheckedConversion(integer, min=1)

def _setting_from_environ(converter, environ_name, default):
    result = default
    env_val = os.environ.get(environ_name, default)
    if env_val is not default:
        try:
            result = converter(env_val)
        except (ValueError, TypeError):
            logger.exception("Failed to parse environment value %r for key %r",
                             env_val, environ_name)
            result = default

    logger.debug('Using value %s from environ %r=%r (default=%r)',
                 result, environ_name, env_val, default)
    return result


def get_positive_integer_from_environ(environ_name, default):
    return _setting_from_environ(positive_integer, environ_name, default)

def parse_boolean(val):
    if val == '0':
        return False
    if val == '1':
        return True
    return asBoolean(val)

parse_byte_size = stock_datatypes['byte-size']

def timestamp_at_unixtime(now):
    """
    Return a :class:`persistent.timestamp.TimeStamp` for the moment
    given by *now* (a float giving seconds since the epoch).
    """

    # ``now`` is a float (double precision.). We extract the seconds
    # value using ``now % 60.0``. But performing a modulus operation
    # on it tends to invent precision that we do not really have.
    #
    # For example, given the timestamp ``1564178539.353595``, the
    # correct value for seconds is obviously ``19.353595``. But Python
    # tends to answer ``19.35359501838684``. When the TimeStamp
    # constructor takes that double precision value and applies *more*
    # arithmetic to it to set the bias, the results get worse: we get
    # ``1385384294.4`` when we should get ``1385384293.0838187``.
    #
    # If we're comparing with a system that doesn't use exactly IEEE double
    # arithmetic to calculate the seconds value, this can lead to wrong answers.

    gmtime = time.gmtime(now)
    seconds = now % 60.0
    return TimeStamp(*(gmtime[:5] + (seconds,)))

class timer(object):
    __begin = None
    __end = None
    duration = None

    counter = perf_counter

    def __enter__(self):
        self.__begin = self.counter()
        return self

    def __exit__(self, t, v, tb):
        self.__end = self.counter()
        self.duration = self.__end - self.__begin

def get_time_from_environ(environ_name, default):
    return _setting_from_environ(float, environ_name, default)

def _get_log_time_level(level_int, default):
    level_name = logging.getLevelName(level_int)
    val = get_time_from_environ('RS_PERF_LOG_%s_MIN' % level_name, default)
    return (level_int, float(val))

# A list of tuples (level_int, min_duration), ordered by increasing
# min_duration. If you modify this list at runtime, be sure to keep it
# sorted. A quick way to completely disable logging is to clear this
# list. Modify this list in place to apply to all functions; make a
# copy and place in `func.__wrapped__.log_levels` or
# `Class.meth.__wrapped__.log_levels` to change for an individual
# function.
_LOG_TIMED_DEFAULT_DURATIONS = [
    _get_log_time_level(TRACE, 0.31),
    _get_log_time_level(DEBUG, 1.24),
    _get_log_time_level(INFO, 3.03),
    _get_log_time_level(WARN, 9.24),
    _get_log_time_level(ERROR, 20.10)
]

_LOG_TIMED_DEFAULT_DURATIONS.sort(key=lambda x: x[1])

# timed events above this threshold will include detail information.
# The 'log_details_threshold' property of the function can be
# assigned to make it different than the default.
_LOG_TIMED_DEFAULT_DETAILS_THRESHOLD = logging.getLevelName(
    _setting_from_environ(str, 'RS_PERF_LOG_DETAILS_LEVEL', 'WARN')
)


# If this is true when a module is imported, timer decorations
# are omitted.
_LOG_TIMED_COMPILETIME_ENABLE = _setting_from_environ(
    parse_boolean,
    'RS_PERF_LOG_ENABLE',
    'on'
)

def do_log_duration_info(basic_msg, func,
                         args, kwargs,
                         actual_duration,
                         log=perf_logger):
    if func is None:
        # Timing was disabled at compile time
        return

    log_level = 0
    # Defer capturing the name; it might get changed at
    # runtime.
    log_msg = basic_msg
    log_args = (func.__name__, actual_duration)
    for level, duration in func.log_levels:
        if actual_duration < duration:
            break
        log_level = level

    if not log.isEnabledFor(log_level):
        return

    if log_level >= func.log_details_threshold:
        # This will capture 'self' as the first argument,
        # so you can put useful things into that repr if you'd
        # like.
        try:
            load = os.getloadavg()
        except (OSError, AttributeError):
            load = "<unknown load>"
        mem = byte_display(get_memory_usage())

        if args and kwargs:
            log_msg += " (load=%s) (memory=%s) (args=%r kwargs=%r)"
            log_args += (load, mem, args, kwargs)
        elif args:
            if getattr(func, 'log_args_only_self', None):
                args = args[:1]
            log_msg += " (load=%s) (memory=%s) (args=%r)"
            log_args += (load, mem, args)

    log.log(log_level, log_msg, *log_args)

def log_timed(func):
    # Store these on each individual function so they can be
    # tweaked later: Class.func.__wrapped__.min_duration_to_log = X
    # Do this even if we've disabled the wrappers for a consistent
    # interface.
    func.log_levels = _LOG_TIMED_DEFAULT_DURATIONS
    func.log_details_threshold = _LOG_TIMED_DEFAULT_DETAILS_THRESHOLD
    if not _LOG_TIMED_COMPILETIME_ENABLE:
        # One more we're possibly missing
        if getattr(func, '__wrapped__', func) is func:
            func.__wrapped__ = None
        return func

    counter = perf_counter
    log = do_log_duration_info
    func_logger = logging.getLogger(func.__module__).getChild('timing')


    @wraps(func)
    def f(*args, **kwargs):
        begin = counter()
        try:
            result = func(*args, **kwargs)
        finally:
            end = counter()
            duration = end - begin

            log_msg = "Function %s took %.3fs."
            log(log_msg, func, args, kwargs, duration,
                log=func_logger)

        return result

    return f


def log_timed_only_self(func):
    func.log_args_only_self = 1
    return log_timed(func)

_ThreadWithReady = None

def thread_spawn(func, args=(), daemon=False):
    global _ThreadWithReady
    if _ThreadWithReady is None:
        import threading

        class T(threading.Thread):
            # Note that this is going to be a bit racy
            # between threads; it may not appear in a reading
            # thread right away.
            __ready = False

            def ready(self):
                return self.__ready

            def run(self):
                try:
                    super(T, self).run()
                finally:
                    self.__ready = True

            def wait(self, timeout=None):
                return self.join(timeout)

        _ThreadWithReady = T

    t = _ThreadWithReady(target=func, args=args)
    t.name = t.name + '-spawn-' + func.__name__
    if daemon:
        t.setDaemon(daemon)
    t.start()
    return t

def _gevent_pool_spawn(func, args=()):
    import gevent
    return gevent.get_hub().threadpool.spawn(func, *args)

def spawn(func, args=()):
    """
    Execute func in a different (real) thread.

    Returns an object with a ``ready()`` method that can be called to
    determine if the *func* has finished running, and a
    ``wait([timeout])`` method that blocks until the function has
    finished running. (This abstracts differences between the gevent
    threadpool and direct use of threads.)
    """

    submit = thread_spawn
    try:
        import gevent.monkey
    except ImportError: # pragma: no cover
        pass
    else:
        if gevent.monkey.is_module_patched('threading'):
            submit = _gevent_pool_spawn

    logger.debug("Using %s to run %s", submit, func)
    return submit(func, args)

def get_this_psutil_process():
    # Returns a freshly queried object each time.
    try:
        from psutil import Process, AccessDenied
        # Make sure it works (why would we be denied access to our own process?)
        try:
            proc = Process()
            proc.memory_full_info()
        except AccessDenied: # pragma: no cover
            proc = None
    except ImportError: # pragma: no cover
        proc = None
    return proc

def get_memory_usage():
    """
    Get a number representing the current memory usage.

    Returns 0 if this is not available.
    """
    proc = get_this_psutil_process()
    if proc is None: # pragma: no cover
        return 0

    rusage = proc.memory_full_info()
    # uss only documented available on Windows, Linux, and OS X.
    # If not available, fall back to rss as an aproximation.
    mem_usage = getattr(rusage, 'uss', 0) or rusage.rss
    return mem_usage

def byte_display(size):
    """
    Returns a string with the correct unit (KB, MB), given the size in bytes.
    """
    if size == 0:
        return '0 KB'
    if size <= 1024:
        return '%s bytes' % size
    if size > 1048576:
        return '%0.02f MB' % (size / 1048576.0)
    return '%0.02f KB' % (size / 1024.0)

class Lazy(object):
    "Property-like descriptor that calls func only once per instance."

    # Derived from zope.cachedescriptors.property.Lazy

    def __init__(self, func, name=None):
        if name is None:
            name = func.__name__
        self.data = (func, name)
        update_wrapper(self, func)

    def __get__(self, inst, class_):
        if inst is None:
            return self

        func, name = self.data
        value = func(inst)
        inst.__dict__[name] = value
        self._stored_value_for_name_in_inst(value, name, inst)
        return value

    @staticmethod
    def _stored_value_for_name_in_inst(value, name, inst):
        """
        Hook for subclasses.
        """

class CachedIn(object):
    """Cached method with given cache attribute."""

    def __init__(self, attribute_name, factory=dict):
        self.attribute_name = attribute_name
        self.factory = factory

    def __call__(self, func):

        @wraps(func)
        def decorated(instance):
            cache = self.cache(instance)
            key = () # We don't support arguments right now, so only one key.
            try:
                v = cache[key]
            except KeyError:
                v = cache[key] = func(instance)
            return v

        decorated.invalidate = self.invalidate
        return decorated

    def invalidate(self, instance):
        cache = self.cache(instance)
        key = ()
        try:
            del cache[key]
        except KeyError:
            pass

    def cache(self, instance):
        try:
            cache = getattr(instance, self.attribute_name)
        except AttributeError:
            cache = self.factory()
            setattr(instance, self.attribute_name, cache)
        return cache


def to_utf8(data):
    if data is None or isinstance(data, bytes):
        return data
    return data.encode("utf-8")

def consume(iterator, n=None):
    "Advance the iterator n-steps ahead. If n is none, consume entirely."
    # From the itertools documentation.
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        collections.deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(itertools.islice(iterator, n, n), None)

class CloseTracker(object): # pragma: no cover
    __slots__ = ('_tracked', '_type', '_at')

    def __init__(self, conn):
        self._tracked = conn
        self._type = type(conn)
        self._at = ''.join(traceback.format_stack())

    def __getattr__(self, name):
        return getattr(self._tracked, name)

    def __setattr__(self, name, value):
        if name in CloseTracker.__slots__:
            object.__setattr__(self, name, value)
            return
        return setattr(self._tracked, name, value)

    def close(self):
        if self._tracked is None:
            return
        try:
            self._tracked.close()
        finally:
            self._tracked = None

    def __del__(self):
        if self._tracked is not None:
            print("Failed to close", self, self._type, " from:", self._at, file=sys.stderr)
            print("Deleted at", ''.join(traceback.format_stack()))


class CloseTrackedConnection(CloseTracker): # pragma: no cover
    __slots__ = ()

    def cursor(self, *args, **kwargs):
        return CloseTrackerCursor(self._tracked.cursor(*args, **kwargs))


class CloseTrackerCursor(CloseTracker):
    __slots__ = ()

    def execute(self, stmt, args=None):
        return self._tracked.execute(stmt, args)

    def __iter__(self):
        return self._tracked.__iter__()


class NeedsFetchallBeforeCloseCursor(CloseTrackerCursor): # pragma: no cover
    needs_fetchall = False
    verbose = False
    last_stmt = None

    def close(self):
        if self.needs_fetchall:
            raise AssertionError("Forgot to fetchall")
        super(NeedsFetchallBeforeCloseCursor, self).close()

    def execute(self, stmt, args=None):
        if self.needs_fetchall:
            raise AssertionError("Forgot to fetchall")
        self.last_stmt = stmt
        if stmt.strip().startswith('SELECT'):
            self.needs_fetchall = True
        if self.verbose:
            print(stmt, end=' -> ')
            result = self._tracked.execute(stmt, args)
            print(result)
            return result
        return self._tracked.execute(stmt, args)

    def fetchone(self):
        r = self._tracked.fetchone()
        if r is None:
            self.needs_fetchall = False
        return r

    def fetchall(self):
        self.needs_fetchall = False
        return self._tracked.fetchall()

    def __repr__(self):
        return "<%r need_fetch=%s last=%r>" % (
            self._tracked,
            self.needs_fetchall,
            self.last_stmt,
        )



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Database schema installers
"""
from __future__ import absolute_import

import re

from zope.interface import implementer

from ..interfaces import ISchemaInstaller
from ..schema import AbstractSchemaInstaller

# Versions of the installed stored procedures. Change these when
# the corresponding code changes.
oracle_package_version = '1.5A'

oracle_history_preserving_package = """
CREATE OR REPLACE PACKAGE relstorage_op AS
    TYPE numlist IS TABLE OF NUMBER(20) INDEX BY BINARY_INTEGER;
    TYPE md5list IS TABLE OF VARCHAR2(32) INDEX BY BINARY_INTEGER;
    TYPE statelist IS TABLE OF RAW(2000) INDEX BY BINARY_INTEGER;
    PROCEDURE store_temp(
        zoids IN numlist,
        prev_tids IN numlist,
        md5s IN md5list,
        states IN statelist);
    PROCEDURE restore(
        zoids IN numlist,
        tids IN numlist,
        md5s IN md5list,
        states IN statelist);
END relstorage_op;
/

CREATE OR REPLACE PACKAGE BODY relstorage_op AS
/* Version: %(oracle_package_version)s */
    PROCEDURE store_temp(
        zoids IN numlist,
        prev_tids IN numlist,
        md5s IN md5list,
        states IN statelist) IS
    BEGIN
        FORALL indx IN zoids.first..zoids.last
            DELETE FROM temp_store WHERE zoid = zoids(indx);
        FORALL indx IN zoids.first..zoids.last
            INSERT INTO temp_store (zoid, prev_tid, md5, state) VALUES
            (zoids(indx), prev_tids(indx), md5s(indx), states(indx));
    END store_temp;

    PROCEDURE restore(
        zoids IN numlist,
        tids IN numlist,
        md5s IN md5list,
        states IN statelist) IS
    BEGIN
        FORALL indx IN zoids.first..zoids.last
            DELETE FROM object_state
            WHERE zoid = zoids(indx)
                AND tid = tids(indx);
        FORALL indx IN zoids.first..zoids.last
            INSERT INTO object_state
                (zoid, tid, prev_tid, md5, state_size, state)
                VALUES (zoids(indx), tids(indx),
                    COALESCE((SELECT tid
                        FROM current_object
                        WHERE zoid = zoids(indx)), 0),
                md5s(indx), COALESCE(LENGTH(states(indx)), 0), states(indx));
    END restore;
END relstorage_op;
/
""" % globals()


oracle_history_free_package = """
CREATE OR REPLACE PACKAGE relstorage_op AS
    TYPE numlist IS TABLE OF NUMBER(20) INDEX BY BINARY_INTEGER;
    TYPE md5list IS TABLE OF VARCHAR2(32) INDEX BY BINARY_INTEGER;
    TYPE statelist IS TABLE OF RAW(2000) INDEX BY BINARY_INTEGER;
    PROCEDURE store_temp(
        zoids IN numlist,
        prev_tids IN numlist,
        md5s IN md5list,
        states IN statelist);
    PROCEDURE restore(
        zoids IN numlist,
        tids IN numlist,
        states IN statelist);
END relstorage_op;
/

CREATE OR REPLACE PACKAGE BODY relstorage_op AS
/* Version: %(oracle_package_version)s */
    PROCEDURE store_temp(
        zoids IN numlist,
        prev_tids IN numlist,
        md5s IN md5list,
        states IN statelist) IS
    BEGIN
        FORALL indx IN zoids.first..zoids.last
            DELETE FROM temp_store WHERE zoid = zoids(indx);
        FORALL indx IN zoids.first..zoids.last
            INSERT INTO temp_store (zoid, prev_tid, md5, state) VALUES
            (zoids(indx), prev_tids(indx), md5s(indx), states(indx));
    END store_temp;

    PROCEDURE restore(
        zoids IN numlist,
        tids IN numlist,
        states IN statelist) IS
    BEGIN
        FORALL indx IN zoids.first..zoids.last
            DELETE FROM object_state WHERE zoid = zoids(indx);
        FORALL indx IN zoids.first..zoids.last
            INSERT INTO object_state (zoid, tid, state_size, state) VALUES (
                zoids(indx),
                tids(indx),
                COALESCE(LENGTH(states(indx)), 0),
                states(indx));
    END restore;
END relstorage_op;
/
""" % globals()



@implementer(ISchemaInstaller)
class OracleSchemaInstaller(AbstractSchemaInstaller):

    all_tables = list(AbstractSchemaInstaller.all_tables)
    all_tables.remove('new_oid')
    all_tables = tuple(all_tables)

    all_sequences = (
        'zoid_seq',
    )

    database_type = 'oracle'

    COLTYPE_OID_TID = 'NUMBER(20)'
    COLTYPE_BINARY_STRING = 'RAW(2000)'
    COLTYPE_BLOB_CHUNK = 'BLOB'
    COLTYPE_BLOB_CHUNK_NUM = COLTYPE_OID_TID
    COLTYPE_STATE_SIZE = COLTYPE_OID_TID
    COLTYPE_STATE = 'BLOB'

    def get_database_name(self, cursor):
        cursor.execute("SELECT ora_database_name FROM DUAL")
        for (name,) in cursor:
            return name

    def create_procedures(self, cursor):
        packages = self.list_packages(cursor)
        package_name = 'relstorage_op'
        if packages.get(package_name) != oracle_package_version:
            self.install_package(cursor)
            packages = self.list_packages(cursor)
            if packages.get(package_name) != oracle_package_version:
                raise AssertionError(
                    "Could not get version information after "
                    "installing the %s package." % package_name)

    def install_package(self, cursor):
        """Install the package containing stored procedures"""
        if self.keep_history:
            code = oracle_history_preserving_package
        else:
            code = oracle_history_free_package

        for stmt in code.split('\n/\n'):
            if stmt.strip():
                cursor.execute(stmt)

    def list_tables(self, cursor):
        cursor.execute("SELECT table_name FROM user_tables")
        return [name for (name,) in cursor.fetchall()]

    def list_views(self, cursor):
        cursor.execute("SELECT view_name FROM user_views")
        return [name for (name,) in cursor.fetchall()]

    def list_sequences(self, cursor):
        cursor.execute("SELECT sequence_name FROM user_sequences")
        return {name for (name,) in cursor.fetchall()}

    def list_packages(self, cursor):
        """List installed stored procedure packages.

        Returns {package name: version}.  version may be None.
        """
        stmt = """
        SELECT object_name
        FROM user_objects
        WHERE object_type = 'PACKAGE'
        """
        cursor.execute(stmt)
        names = [name for (name,) in cursor.fetchall()]

        res = {}
        for name in names:
            version = None
            stmt = """
            SELECT TEXT FROM USER_SOURCE
            WHERE TYPE='PACKAGE BODY'
                AND NAME=:1
            """
            cursor.execute(stmt, (name,))
            for (text,) in cursor:
                match = re.search(r'Version:\s*([0-9a-zA-Z.]+)', text)
                if match is not None:
                    version = match.group(1)
                    break
            res[name.lower()] = version
        return res

    list_procedures = list_packages

    def _create_pack_lock(self, cursor):
        stmt = "CREATE TABLE pack_lock (dummy CHAR);"
        self.runner.run_script(cursor, stmt)

    def _create_transaction(self, cursor):
        if self.keep_history:
            stmt = """
            CREATE TABLE transaction (
                tid         NUMBER(20) NOT NULL PRIMARY KEY,
                packed      CHAR DEFAULT 'N' CHECK (packed IN ('N', 'Y')),
                is_empty    CHAR DEFAULT 'N' CHECK (is_empty IN ('N', 'Y')),
                username    RAW(500),
                description RAW(2000),
                extension   RAW(2000)
            );
            """
            self.runner.run_script(cursor, stmt)

    def _create_zoid_seq(self, cursor):
        stmt = """
        CREATE SEQUENCE zoid_seq;
        """
        self.runner.run_script(cursor, stmt)

    def _create_blob_chunk(self, cursor):
        if self.keep_history:
            stmt = """
            CREATE TABLE blob_chunk (
                zoid        NUMBER(20) NOT NULL,
                tid         NUMBER(20) NOT NULL,
                chunk_num   NUMBER(20) NOT NULL,
                            PRIMARY KEY (zoid, tid, chunk_num),
                chunk       BLOB
            );
            CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid, tid);
            ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
                FOREIGN KEY (zoid, tid)
                REFERENCES object_state (zoid, tid)
                ON DELETE CASCADE;
            """
        else:
            stmt = """
            CREATE TABLE blob_chunk (
                zoid        NUMBER(20) NOT NULL,
                chunk_num   NUMBER(20) NOT NULL,
                            PRIMARY KEY (zoid, chunk_num),
                tid         NUMBER(20) NOT NULL,
                chunk       BLOB
            );
            CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid);
            ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
                FOREIGN KEY (zoid)
                REFERENCES object_state (zoid)
                ON DELETE CASCADE;
            """

        self.runner.run_script(cursor, stmt)

    def _create_pack_state(self, cursor):
        if self.keep_history:
            stmt = """
            CREATE TABLE pack_state (
                tid         NUMBER(20) NOT NULL,
                zoid        NUMBER(20) NOT NULL,
                PRIMARY KEY (tid, zoid)
            );
            """
            self.runner.run_script(cursor, stmt)

    def _create_pack_state_tid(self, cursor):
        if self.keep_history:
            stmt = """
            CREATE TABLE pack_state_tid (
                tid         NUMBER(20) NOT NULL PRIMARY KEY
            );
            """
            self.runner.run_script(cursor, stmt)

    def _reset_oid(self, cursor):
        stmt = """
        DROP SEQUENCE zoid_seq;
        CREATE SEQUENCE zoid_seq;
        """
        self.runner.run_script(cursor, stmt)

    def _create_temp_store(self, cursor):
        stmt = """
            CREATE GLOBAL TEMPORARY TABLE temp_store (
                zoid        NUMBER(20) NOT NULL PRIMARY KEY,
                prev_tid    NUMBER(20) NOT NULL,
                md5         CHAR(32),
                state       BLOB,
                blobdata    BLOB
            ) ON COMMIT DELETE ROWS;
            """
        self.runner.run_script(cursor, stmt)

    def _create_temp_blob_chunk(self, cursor):
        stmt = """
            CREATE GLOBAL TEMPORARY TABLE temp_blob_chunk (
                zoid        NUMBER(20) NOT NULL,
                chunk_num   NUMBER(20) NOT NULL,
                chunk       BLOB
            ) ON COMMIT DELETE ROWS;
            """
        self.runner.run_script(cursor, stmt)


    def _create_temp_pack_visit(self, cursor):
        stmt = """
            CREATE GLOBAL TEMPORARY TABLE temp_pack_visit (
                zoid        NUMBER(20) NOT NULL PRIMARY KEY,
                keep_tid    NUMBER(20) NOT NULL
            );

            """
        self.runner.run_script(cursor, stmt)

    def _create_temp_undo(self, cursor):
        if self.keep_history:
            stmt = """
            CREATE GLOBAL TEMPORARY TABLE temp_undo (
                zoid        NUMBER(20) NOT NULL PRIMARY KEY,
                prev_tid    NUMBER(20) NOT NULL
            );

            """
            self.runner.run_script(cursor, stmt)



##############################################################################
#
# Copyright (c) 2009,2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import os
import ZODB.blob

from ZODB.utils import p64

from zope.interface import implementer

from relstorage.interfaces import POSKeyError
from .interfaces import IAuthoritativeBlobHelper
from .abstract import AbstractBlobHelper


logger = __import__('logging').getLogger(__name__)


@implementer(IAuthoritativeBlobHelper)
class SharedBlobHelper(AbstractBlobHelper):

    NEEDS_DB_LOCK_TO_VOTE = True
    NEEDS_DB_LOCK_TO_FINISH = False

    def __init__(self, options, adapter, fshelper=None):
        assert options.shared_blob_dir

        if fshelper is None:
            # Share files over NFS or similar
            fshelper = ZODB.blob.FilesystemHelper(options.blob_dir)
            fshelper.create()
        super(SharedBlobHelper, self).__init__(options, adapter, fshelper)

    def _loadBlobInternal(self, cursor, oid, serial, blob_lock=None):
        blob_filename = self._cachedLoadBlobInternal(oid, serial)
        if not blob_filename:
            # All the blobs are in a shared directory. If the file
            # isn't here, it's not anywhere.
            raise POSKeyError(oid, serial=serial, blob_filename=blob_filename)
        return blob_filename

    def storeBlob(self, _cursor, store_func,
                  oid, serial, data, blobfilename, version, txn):
        assert not version
        self._doStoreBlob(store_func, oid, serial, data, blobfilename, txn)

    def restoreBlob(self, _cursor, oid, serial, blobfilename):
        self.fshelper.getPathForOID(oid, create=True)
        targetname = self.fshelper.getBlobFilename(oid, serial)
        ZODB.blob.rename_or_copy_blob(blobfilename, targetname)

    def copy_undone(self, copied, tid):
        """
        After an undo operation, copy the matching blobs forward.

        The copied parameter is a list of ``(oid_int, tid_int)``.
        """
        for oid_int, old_tid_int in copied:
            oid = p64(oid_int)
            old_tid = p64(old_tid_int)
            orig_fn = self.fshelper.getBlobFilename(oid, old_tid)
            if not os.path.exists(orig_fn):
                # not a blob
                continue

            new_fn = self.fshelper.getBlobFilename(oid, tid)
            with open(orig_fn, 'rb') as orig, open(new_fn, 'wb') as new:
                ZODB.utils.cp(orig, new)

            self._add_blob_to_transaction(oid, new_fn)

    @staticmethod
    def _has_files(dirname):
        """Return True if a directory has any visible files."""
        names = os.listdir(dirname)
        if not names:
            return False
        for name in names:
            if not name.startswith('.'):
                return True
        return False

    def after_pack(self, oid_int, tid_int):
        """
        Called after an object state has been removed by packing.

        Removes the corresponding blob file.
        """
        oid = p64(oid_int)
        tid = p64(tid_int)
        fn = self.fshelper.getBlobFilename(oid, tid)
        if self.adapter.keep_history:
            # remove only the revision just packed
            if os.path.exists(fn):
                ZODB.blob.remove_committed(fn)
                dirname = os.path.dirname(fn)
                if not self._has_files(dirname):
                    ZODB.blob.remove_committed_dir(dirname)
        else:
            # remove all revisions
            dirname = os.path.dirname(fn)
            if os.path.exists(dirname):
                for name in os.listdir(dirname):
                    ZODB.blob.remove_committed(os.path.join(dirname, name))
                ZODB.blob.remove_committed_dir(dirname)

    def vote(self, tid=None):
        self._move_blobs_into_place(tid)

    def _abort_filename(self, filename):
        dirname = os.path.dirname(filename)
        if not self._has_files(dirname):
            ZODB.blob.remove_committed_dir(dirname)



# -*- coding: utf-8 -*-
"""
Tests for vote.py.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import unittest

from hamcrest import assert_that
from nti.testing.matchers import verifiably_provides

from ...interfaces import ITPCStateVoting
from .. import SharedTPCState

class MockStoreConnectionPool(object):

    def __init__(self):
        self.conn = object()

    def borrow(self):
        return self.conn

class MockStorage(object):

    def __init__(self):
        self._store_connection_pool = MockStoreConnectionPool()

class MockBeginState(object):

    def __init__(self, shared_state=None):
        self.shared_state = shared_state or SharedTPCState(None, MockStorage(), None)
        self.required_tids = {}
        self.ude = (b'user', b'description', b'extension')
        self.invalidated_oids = set()

class MockLockedBeginState(MockBeginState):

    def __init__(self, shared_state=None):
        from relstorage.tests import MockAdapter
        from ..vote import DatabaseLockedForTid
        MockBeginState.__init__(self, shared_state)
        self.committing_tid_lock = DatabaseLockedForTid(
            b'12345678',
            12345678,
            MockAdapter()
        )

class _InterfaceMixin(object):

    BeginState = MockBeginState

    def _getClass(self):
        raise NotImplementedError

    def _makeOne(self):
        return self._getClass()(self.BeginState())

    def test_provides_interface(self):
        assert_that(self._makeOne(), verifiably_provides(ITPCStateVoting))



class TestHistoryFree(_InterfaceMixin, unittest.TestCase):

    def _getClass(self):
        from ..vote import HistoryFree
        return HistoryFree

class TestHistoryPreserving(_InterfaceMixin, unittest.TestCase):

    BeginState = MockLockedBeginState

    def _getClass(self):
        from ..vote import HistoryPreserving
        return HistoryPreserving

class TestHistoryPreservingDeletOnly(_InterfaceMixin, unittest.TestCase):

    BeginState = MockLockedBeginState

    def _getClass(self):
        from ..vote import HistoryPreservingDeleteOnly
        return HistoryPreservingDeleteOnly
